import torch
import torch.nn as nn

from nnts import utils

from .. import datasets


class SegLSTMDecoder(nn.Module):

    def __init__(self, params: utils.Hyperparams, input_dim: int, hidden_dim: int):
        super(SegLSTMDecoder, self).__init__()
        self.params = params
        self.dim = input_dim
        self.hidden_dim = hidden_dim
        self.rnn = nn.LSTM(
            self.dim,
            self.hidden_dim,
            params.n_layers,
            dropout=params.dropout,
            batch_first=True,
        )
        self.rnn_type = params.rnn_type

    def init_hidden_zeros(self, batch_size):
        if self.rnn_type == "gru":
            hidden = torch.zeros(
                self.params.n_layers,
                batch_size,
                self.hidden_dim,
            )
        else:
            h0 = torch.zeros(
                self.params.n_layers,
                batch_size,
                self.hidden_dim,
            )
            c0 = torch.zeros(
                self.params.n_layers,
                batch_size,
                self.hidden_dim,
            )
            hidden = (h0, c0)
        return hidden

    def forward(self, X: torch.tensor, hidden: bool = None):
        """
        X: B, T, T_dim, C
         - B: batch size
         - T: time steps
         - T_dim: segment length
         - C: covariate length + 1
        """
        B, T, T_dim, C = X.shape
        X = X.reshape(B, T, -1)
        if hidden is None:
            hidden = self.init_hidden_zeros(B)
        out, hidden = self.rnn(X, hidden)
        return out


class SegLSTM(nn.Module):

    def __init__(
        self,
        Distribution: nn.Module,
        params: utils.Hyperparams,
        scaling_fn: callable,
        output_dim: int,
        segment_length: int,
    ):
        """
        Initialize the SegLSTM model.

        Args:
            Distribution (nn.Module): The distribution module used for generating output.
            params (utils.Hyperparams): Hyperparameters for the model.
            scaling_fn (callable): A scaling function applied to the output.
            output_dim (int): The dimension of the output. (ie y + covariates).
            segment_length (int): The segment length (eg seasonality).

        """
        super(SegLSTM, self).__init__()
        self.decoder = SegLSTMDecoder(params, segment_length * output_dim, 40)
        self.distribution = Distribution(40, segment_length * output_dim)
        self.output_dim = output_dim
        self.scaling_fn = scaling_fn
        self.segment_length = segment_length

    def forward(self, X: torch.tensor, pad_mask: torch.tensor) -> torch.tensor:

        X = X.clone()
        B, T, C = X.shape

        if self.scaling_fn is None:
            target_scale = None
            x = X
        else:
            target_scale = self.scaling_fn(X, pad_mask)
            x = X / target_scale

        x = x.reshape(B, -1, self.segment_length, C)  # B, T/12, 12, C
        B, T, T_dim, C = x.shape

        x = self.decoder(x)
        x = self.distribution(x, None)
        x = x.reshape(B, T, T_dim, C)
        x = x * target_scale.unsqueeze(2)
        return x

    def generate(
        self,
        X: torch.tensor,
        pad_mask: torch.tensor,
        prediction_length: int,
        context_length: int,
    ) -> torch.tensor:
        pred_list = []
        while True:
            pad_mask = pad_mask[:, -context_length + 1 :]
            preds = self.forward(X[:, -context_length + 1 :, :], pad_mask)
            # focus only on the last time step
            preds = preds[:, -1:, -1, :]
            pred_list.append(preds)

            if len(pred_list) >= prediction_length:
                break
            y_hat = preds.detach().clone()
            X = torch.cat((X, y_hat), dim=1)  # (B, T+1)
            pad_mask = torch.cat((pad_mask, torch.ones_like(preds[:, :, 0])), dim=1)
        return torch.cat(pred_list, 1)

    def teacher_forcing_output(self, batch: datasets.PaddedData, *args, **kwargs):
        x = batch.data
        B, T, C = x.shape
        pad_mask = batch.pad_mask
        y_hat = self(x[:, :-1, :], pad_mask[:, :-1])
        y_hat = y_hat[:, :, -1:, ...]  # B, T, S, C
        # y_hat = y_hat[:, :, :, ...]  # B, T, S, C

        y = x[:, 1:, :]
        y = y.reshape(B, -1, self.segment_length, C)
        y = y[:, :, -1:, ...]  # B, T, S, C
        # y = y[:, :, :, ...]  # B, T, S, C
        return y_hat, y



import json
import os
import warnings
from collections import namedtuple
from typing import Tuple

import pandas as pd
from pandas.errors import OutOfBoundsDatetime
from pydantic import BaseModel

import nnts.data.tsf as tsf

warnings.simplefilter(action="ignore", category=FutureWarning)

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")
SplitData = namedtuple("SplitData", ["train", "validation", "test"])
SplitTrainTest = namedtuple("SplitTrainTest", ["train", "test"])

FREQUENCY_MAP: dict = {
    "minutely": "1min",
    "10_minutes": "10min",
    "half_hourly": "30min",
    "hourly": "1H",
    "daily": "1D",
    "weekly": "1W",
    "monthly": "1M",
    "quarterly": "1Q",
    "yearly": "1Y",
}


class Metadata(BaseModel):
    """Class for storing metadata information about a dataset.

    Attributes:
    filename (str): The name of the file containing the dataset.
    dataset (str): The name of the dataset.
    context_length (int): The number of past observations used as context for making predictions.
    prediction_length (int): The number of future observations to predict.
    freq (str): The frequency of the time series data (e.g., 'D' for daily, 'H' for hourly).
    seasonality (int): The seasonal period of the time series data.
    url (str, optional): The URL where the dataset can be accessed. Defaults to None.
    """

    filename: str
    dataset: str
    context_length: int
    prediction_length: int
    freq: str
    seasonality: int
    url: str = None
    context_lengths: list = None
    multivariate: bool = False


def unpack(df: pd.DataFrame, freq: str = "ME") -> pd.DataFrame:
    try:
        timesteps = pd.date_range(
            df["start_timestamp"], periods=len(df["series_value"]), freq=freq
        )
    except OutOfBoundsDatetime:
        timesteps = pd.date_range(
            pd.Timestamp("1970-01-01"),
            periods=len(df["series_value"]),
            freq=freq,
            unit="s",
        )

    unpacked_df = pd.DataFrame(
        data={"y": df["series_value"].to_numpy(), "ds": timesteps.astype(str)}
    )
    unpacked_df["unique_id"] = df["series_name"]
    return unpacked_df


def read_tsf(path: str, url: str = None) -> pd.DataFrame:
    """Reads a time series forecasting (TSF) file from a specified path or URL and returns
    the data as a pandas DataFrame along with metadata.

    Parameters
    ----------
    path: str
        The file path to the TSF file.
    url: :obj: `str` optional
        The URL to the TSF file. If specified, the file will be downloaded from the URL. Defaults to None.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the time series data with the necessary transformations applied.
    str
        The frequency of the time series data.
    int
        The forecast horizon.
    bool
        Indicator whether the data contains missing values.
    bool
        Indicator whether the time series data have equal lengths.

    Example
    -------
    >>> df, freq, forecast_horizon, contain_missing_values, contain_equal_length = read_tsf('path/to/file.tsf')
    >>> print(df.head())
    >>> print(freq)
    >>> print(forecast_horizon)
    >>> print(contain_missing_values)
    >>> print(contain_equal_length)
    """
    (
        all_data,
        frequency,
        forecast_horizon,
        contain_missing_values,
        contain_equal_length,
    ) = tsf.read(path, url)
    df = pd.DataFrame(all_data)
    freq = FREQUENCY_MAP[frequency]
    if "start_timestamp" not in df.columns:
        df["start_timestamp"] = pd.Timestamp("1970-01-01")
    df = pd.concat([unpack(df.iloc[x], freq=freq) for x in range(len(df))])
    return df, freq, forecast_horizon, contain_missing_values, contain_equal_length


def load_metadata(
    dataset: str,
    path: str = None,
) -> Metadata:
    """
    Loads metadata for a specified dataset from a JSON file.

    Parameters
    ----------
    dataset: str
        The name of the dataset for which to load metadata.
    path: :obj: `str` optional
        The path to the JSON file containing metadata. If not specified, a default path should be provided by the caller.

    Returns
    -------
    Metadata
        An instance of the Metadata class initialized with the data corresponding to the specified dataset.

    Raises
    ------
    ValueError
        If the specified dataset is not found in the metadata JSON file.

    Example
    -------
    >>> metadata = load_metadata('my_dataset', 'path/to/metadata.json')
    >>> print(metadata)
    Metadata(attribute1=value1, attribute2=value2, ...)
    """
    with open(path) as f:
        data = json.load(f)
    if dataset not in data:
        raise ValueError(
            f"Dataset {dataset} not found in metadata {path} choose from: {', '.join(data.keys())}"
        )
    return Metadata(**data[dataset])


def load_dataset(
    name: str, repository: str = "monash"
) -> Tuple[pd.DataFrame, Metadata]:
    """
    Load dataset and its metadata.

    Parameters
    ----------
    name: str
        Name of the dataset to load.
    repository: str
        Name of the repository where the dataset metadata is stored. Default is 'monash'.

    Returns
    -------
    Tuple[pd.DataFrame, Metadata]
        A tuple containing the loaded DataFrame and its metadata.
    """
    path = os.path.join(DATA_PATH, f"{repository}.json")
    metadata = load_metadata(name, path=path)

    df, *_ = read_tsf(metadata.filename, metadata.url)
    return df, metadata


def load(
    dataset_name: str, data_path: str, metadata_filename: str
) -> Tuple[pd.DataFrame, Metadata]:

    metadata_path = os.path.join(data_path, metadata_filename)
    metadata = load_metadata(dataset_name, path=metadata_path)
    datafile_path = os.path.join(data_path, metadata.filename)
    df, freq, forecast_horizon, *_ = read_tsf(datafile_path)
    metadata.freq = freq
    if forecast_horizon is not None:
        metadata.prediction_length = forecast_horizon
    return df, metadata


def split_test_train_last_horizon(
    data: pd.DataFrame, context_length: int, prediction_length: int
) -> SplitTrainTest:
    """
    Splits the given time series data into training and testing sets based on the
    specified context length and prediction length.

    Parameters
    ----------
    data: pd.DataFrame
        The input time series data containing a 'unique_id' column to identify different series.
    context_length: int
        The number of past observations to use as context for making predictions.
    prediction_length: int
        The number of future observations to predict.

    Returns
    -------
    SplitTrainTest
        A named tuple containing two DataFrames:
            - train: The training set, which excludes the last 'prediction_length' observations for each unique series.
            - test: The testing set, which includes the last 'context_length + prediction_length' observations for each unique series.

    """
    trn = data.groupby("unique_id").head(-prediction_length)
    test = data.groupby("unique_id").tail(context_length + prediction_length)
    return SplitTrainTest(train=trn, test=test)


def split_test_val_train_last_horizon(
    data: pd.DataFrame, context_length: int, prediction_length: int
) -> SplitTrainTest:
    """
    Splits the given time series data into training, validation, and testing sets based
    on the specified context length and prediction length.

    Parameters
    ----------
    data: pd.DataFrame
        The input time series data containing a 'unique_id' column to identify different series.
    context_length: int
        The number of past observations to use as context for making predictions.
    prediction_length: int
        The number of future observations to predict.

    Returns
    -------
    SplitData
        A named tuple containing three DataFrames:
            - train: The training set, which excludes the last '2 * prediction_length' observations for each unique series.
            - validation: The validation set, which excludes the last 'prediction_length' and includes a time-range of context_length + prediction_length or each unique series.
            - test: The testing set, which includes the last 'context_length + prediction_length' observations for each unique series.
    """
    trn_val = data.groupby("unique_id").head(-prediction_length)
    trn = trn_val.groupby("unique_id").head(-prediction_length)
    val = trn_val.groupby("unique_id").tail(context_length + prediction_length)
    test = data.groupby("unique_id").tail(context_length + prediction_length)
    return SplitData(train=trn, validation=val, test=test)


def split_test_val_train(
    data: pd.DataFrame,
    trn_length: int,
    val_length: int,
    test_length: int,
    context_length: int = 0,
    prediction_length: int = 336,
) -> SplitTrainTest:
    data = data.groupby("unique_id").head(trn_length + val_length + test_length)
    trn_val = data.groupby("unique_id").head(trn_length + val_length - context_length)
    trn = trn_val.groupby("unique_id").head(trn_length)
    val = trn_val.groupby("unique_id").tail(val_length + prediction_length)
    test = data.groupby("unique_id").tail(test_length + prediction_length)
    return SplitData(train=trn, validation=val, test=test)



import json
import os
import shutil
import timeit
import types
from abc import ABC, abstractmethod
from enum import Enum
from functools import singledispatchmethod
from typing import Any, Dict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import wandb

from . import events, trainers, utils


class Handler(ABC):

    @abstractmethod
    def handle(self, data: Any) -> None:
        pass


class PrintHandler(Handler):
    def handle(self, data: Dict[str, Any]) -> None:
        print(data)


def convert_np_float(obj):
    if isinstance(obj, np.float32):
        return float(obj)
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, type):
        return obj.__name__
    if isinstance(obj, types.FunctionType):
        return obj.__name__
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


class JsonFileHandler(Handler):
    def __init__(self, path, filename):
        self.path = path
        self.filename = filename + ".json"

    def handle(self, data: Dict[str, Any]) -> None:
        with open(os.path.join(self.path, self.filename), "w") as json_file:
            json.dump(data, json_file, indent=4, default=convert_np_float)


class EpochEventMixin(events.Listener):

    @singledispatchmethod
    def notify(self, event: Dict[str, Any]) -> None:
        self.log(event)

    @notify.register(trainers.EpochTrainComplete)
    def _(self, event: trainers.EpochTrainComplete) -> None:
        self.log(
            {
                "epoch": event.state.epoch,
                "train_loss": event.state.train_loss,  # event.state.train_loss.detach().item(),
            }
        )

    @notify.register(trainers.EpochValidateComplete)
    def _(self, event: trainers.EpochValidateComplete) -> None:
        self.log(
            {
                "valid_loss": event.state.valid_loss  # event.state.valid_loss.detach().item(),
            }
        )
        print(
            f"Epoch {event.state.epoch} train loss: {event.state.train_loss}, valid loss: {event.state.valid_loss}"
        )

    @notify.register(trainers.EpochBestModel)
    def _(self, event: trainers.EpochBestModel) -> None:
        self.log_model(event.path)

    def configure(self, evts: events.EventManager) -> None:
        evts.add_listener(trainers.EpochTrainComplete, self)
        evts.add_listener(trainers.EpochValidateComplete, self)
        evts.add_listener(trainers.EpochBestModel, self)


class Run(ABC):

    @abstractmethod
    def log(self, data: Any) -> None:
        pass

    @abstractmethod
    def log_model(self, name: str, path: str) -> None:
        pass

    @abstractmethod
    def log_activations(self, module, input, output) -> None:
        pass


class ActivationVisualizer:
    def __init__(self):
        self.activations = []

    def has_activations(self):
        return len(self.activations) > 0

    def save_heatmap(self, path):
        if len(self.activations) > 0:
            plt.figure(figsize=(10, 6))
            sns.heatmap(self.activations, cmap="coolwarm", linewidths=0.5)
            plt.savefig(path)

    def append(self, output):
        if isinstance(output, tuple):
            output = output[0]
        input_0 = output[0]
        self.activations.append(input_0[:, -1].detach().cpu().numpy())


class LocalFileRun(Run, EpochEventMixin):

    def __init__(
        self,
        project: str,
        name: str,
        config: Dict[str, Any] = None,
        path: str = "",
        Handler: Handler = JsonFileHandler,
    ):
        self.project = project
        self.name = name
        self.static_data = config
        self.path = path
        utils.makedirs_if_not_exists(self.path)
        self.handler = Handler(path=path, filename=name)
        self.start_time = timeit.default_timer()
        self.activation_visualizer = ActivationVisualizer()

    def log(self, data: Any) -> None:
        self.static_data = {**self.static_data, **data}

    def log_model(self, source_file: str) -> None:
        print(f"Artifact saved to {source_file}")
        try:
            shutil.copy(source_file, self.path)
        except shutil.SameFileError:
            pass

    def log_outputs(self, data: Dict[str, Any]) -> None:
        numpy_dict = {key: value for key, value in data.items()}
        # Save each numpy array to a separate text file
        for key, value in numpy_dict.items():
            file_name = key + ".txt"  # You can use the key as the file name
            np.savetxt(
                os.path.join(self.path, file_name), value, fmt="%f"
            )  # Save the numpy array to a text file

    def log_activations(self, module, input, output) -> None:
        self.activation_visualizer.append(output)

    def finish(self) -> None:
        run_time = timeit.default_timer() - self.start_time
        self.static_data["run_time"] = run_time
        self.handler.handle(self.static_data)
        if self.activation_visualizer.has_activations():
            try:
                self.activation_visualizer.save_heatmap(
                    os.path.join(self.path, "activations.png")
                )
            except Exception as e:
                print(f"Error saving activations: {e}")
        print(f"Run {self.name} finished")


class PrintRun(Run, EpochEventMixin):

    def __init__(self, project: str, name: str, config: Dict[str, Any] = None):
        self.project = project
        self.name = name
        self.static_data = config
        self.handler = PrintHandler()
        self.start_time = timeit.default_timer()

    def log(self, data: Any) -> None:
        self.static_data = {**self.static_data, **data}

    def log_model(self, source_file: str) -> None:
        print(f"Artifact saved to {source_file}")

    def log_outputs(self, data: Dict[str, Any]) -> None:
        print(data)

    def log_activations(self, module, input, output) -> None:
        print("input", input)
        print("output", output)

    def finish(self) -> None:
        run_time = timeit.default_timer() - self.start_time
        self.static_data["run_time"] = run_time
        print(self.static_data)
        print(f"Run {self.name} finished")


class WandbRun(Run, EpochEventMixin):

    def __init__(
        self, project: str, name: str, config: Dict[str, Any] = None, path: str = ""
    ):
        self.project = project
        self.name = name
        self.static_data = config
        self.run = wandb.init(
            project=self.project, name=self.name, config=self.static_data
        )
        self.path = path
        utils.makedirs_if_not_exists(self.path)
        self.activation_visualizer = ActivationVisualizer()

    def log(self, data: Any) -> None:
        self.run.log(data)

    def log_model(self, source_file: str) -> None:
        self.run.log_model(name=f"{self.name}-{self.run.id}", path=source_file)

    def log_activations(self, module, input, output) -> None:
        self.activation_visualizer.append(output)

    def finish(self) -> None:
        print(f"Run {self.name} finished")

        if self.activation_visualizer.has_activations():
            try:
                activation_image_path = os.path.join(self.path, "activations.png")
                self.activation_visualizer.save_heatmap(activation_image_path)
                self.run.log({"activations": wandb.Image(activation_image_path)})
            except Exception as e:
                print(f"Error saving activations: {e}")

        self.run.finish()

    def log_table(self, df: pd.DataFrame, metadata: Dict[str, Any]) -> None:
        table = wandb.Table(dataframe=df)
        table_artifact = wandb.Artifact(metadata.dataset, type="dataset")
        table_artifact.add(table, "table")
        self.run.log_artifact(table_artifact)
        self.log({"results": table})


"""
test_table = wandb.Table(dataframe=test_df)
  test_table_artifact = wandb.Artifact(metadata.dataset, type="dataset")
  test_table_artifact.add(test_table, "test_table")

  logger.log(test_metrics)
  logger.log({"results": test_table})

  # and Log as an Artifact to increase the available row limit!
  lxogger.run.log_artifact(test_table_artifact)
"""



import os
from typing import Any, Tuple

import numpy as np
import torch
import torch.nn.functional as F
import torch.utils.data

import nnts.data
import nnts.loggers
import nnts.metrics
import nnts.trainers
from nnts import datasets, utils

from .datasets import PaddedData


class EarlyStopper:
    def __init__(self, patience: int):
        self.patience = patience
        self.min_loss = np.inf
        self.counter = 0

    def early_stopping(self, loss: float) -> bool:
        if loss < self.min_loss:
            self.min_loss = loss
            self.counter = 0
        else:
            self.counter += 1

        if self.counter >= self.patience or np.isnan(loss):
            print("early stopping")
            return True

        return False


def validate(net, batch: PaddedData, prediction_length, context_length):
    if hasattr(net, "validate"):
        return net.validate(batch, prediction_length, context_length)
    else:
        y = batch.data[:, context_length : context_length + prediction_length, ...]
        y_hat = net.generate(
            batch.data[:, :context_length, ...],
            batch.pad_mask[:, :context_length],
            prediction_length=prediction_length,
            context_length=context_length,
        )
        y_hat = y_hat[:, -prediction_length:, ...]
        return y_hat, y


def eval(
    net, dl, prediction_length: int, context_length: int, hooks: Any = None
) -> Tuple[torch.Tensor, torch.Tensor]:
    net.eval()
    with torch.no_grad():
        y_list = []
        y_hat_list = []
        for batch in dl:
            y_hat, y = validate(net, batch, prediction_length, context_length)
            y_list.append(y[:, :, :])
            y_hat_list.append(y_hat[:, :, :])

            # TODO: improve this. Removing hook after first batch to make visualisation work
            if hooks is not None:
                hooks.remove()

        y = torch.cat(y_list, dim=0)
        y_hat = torch.cat(y_hat_list, dim=0)
    return y_hat, y


class TorchEvaluator(nnts.trainers.Evaluator):
    def __init__(self, net: torch.nn.Module):
        self.net = net

    def evaluate(
        self,
        dl: torch.utils.data.DataLoader,
        prediction_length: int,
        context_length: int,
        hooks=None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        return eval(self.net, dl, prediction_length, context_length, hooks=hooks)


class TorchForecaster(nnts.trainers.Forecaster):
    def __init__(self, net: torch.nn.Module):
        self.net = net

    def forecast_batch(
        self, batch: PaddedData, prediction_length: int, context_length: int
    ) -> Any:
        y_hat = self.net.generate(
            batch.data,
            batch.pad_mask,
            prediction_length=prediction_length,
            context_length=context_length,
        )

        y_hat = y_hat[:, -prediction_length:, ...]
        return y_hat

    def forecast(
        self,
        dl: torch.utils.data.DataLoader,
        prediction_length: int,
        context_length: int,
    ) -> Any:
        self.net.eval()
        with torch.no_grad():
            y_hat_list = []
            for batch in dl:
                y_hat = self.forecast_batch(batch, prediction_length, context_length)
                y_hat_list.append(y_hat[:, :, :1])

            y_hat = torch.cat(y_hat_list, dim=0)
        return y_hat


class ValidationTorchEpochTrainer(nnts.trainers.EpochTrainer):

    def __init__(
        self,
        net: torch.nn.Module,
        params: utils.Hyperparams,
        metadata: datasets.Metadata,
        state: nnts.trainers.TrainerState = None,
        model_path: str = "best_model.pt",
    ):
        super().__init__(state, params)
        self.net = net
        self.metadata = metadata
        utils.makedirs_if_not_exists(params.model_file_path)
        self.path = os.path.join(params.model_file_path, model_path)
        self.loss_fn = params.loss_fn
        self.Optimizer = params.optimizer

    def before_train(self, train_dl):
        print(self.net)
        if self.params.batches_per_epoch is None:
            self.params.batches_per_epoch = len(train_dl)
        self.optimizer = self.Optimizer(
            self.net.parameters(),
            lr=self.params.lr,
            weight_decay=self.params.weight_decay,
        )
        if self.params.scheduler == utils.Scheduler.REDUCE_LR_ON_PLATEAU:
            self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer, mode="min", factor=0.5, patience=self.params.patience
            )
        elif self.params.scheduler == utils.Scheduler.STEP_LR:
            self.scheduler = torch.optim.lr_scheduler.StepLR(
                self.optimizer, step_size=1, gamma=0.5
            )
        elif self.params.scheduler == utils.Scheduler.ONE_CYCLE:
            self.scheduler = torch.optim.lr_scheduler.OneCycleLR(
                self.optimizer,
                max_lr=self.params.lr,
                steps_per_epoch=self.params.batches_per_epoch,
                epochs=self.params.epochs,
                pct_start=0.3,
            )
        self.early_stopper = (
            None
            if self.params.early_stopper_patience is None
            else EarlyStopper(patience=self.params.early_stopper_patience)
        )

        self.best_loss = np.inf

    def before_train_epoch(self) -> None:
        self.net.train()

    def before_validate_epoch(self) -> None:
        self.net.eval()

    def after_validate_epoch(self) -> None:
        if self.early_stopper:
            if self.early_stopper.early_stopping(self.state.valid_loss):
                self.state.stop = True

        if self.state.valid_loss < self.state.best_loss:
            print(
                f"Epoch {self.state.epoch} Train Loss: {self.state.train_loss}  Valid Loss: {self.state.valid_loss}"
            )
            print(f"saving model to {self.path}")
            torch.save(self.net.state_dict(), self.path)
            self.state.best_loss = self.state.valid_loss
            self.events.notify(nnts.trainers.EpochBestModel(self.path))

        if self.params.scheduler == utils.Scheduler.REDUCE_LR_ON_PLATEAU:
            self.scheduler.step(self.state.valid_loss)
        elif self.params.scheduler == utils.Scheduler.STEP_LR:
            if self.state.epoch > 1:
                self.scheduler.step()
                print("reducing lr", self.scheduler.get_last_lr())

    def _train_batch(self, i, batch):
        self.optimizer.zero_grad()
        y_hat, y = self.net.train_output(
            batch,
            self.metadata.prediction_length,
            self.metadata.context_length,
            training_method=self.params.training_method,
        )

        L = self.loss_fn(y_hat, y)
        L.backward()
        self.optimizer.step()
        if self.params.scheduler == utils.Scheduler.ONE_CYCLE:
            self.scheduler.step()  # This is required for OneCycleLR
        return L

    def _validate_batch(self, i, batch):
        with torch.no_grad():
            y_hat, y = validate(
                self.net,
                batch,
                self.metadata.prediction_length,
                self.metadata.context_length,
            )
            L = self.loss_fn(y_hat, y)
        return L

    def create_evaluator(self) -> nnts.trainers.Evaluator:
        state_dict = torch.load(self.path)
        self.net.load_state_dict(state_dict)
        return TorchEvaluator(self.net)


class TorchEpochTrainer(nnts.trainers.EpochTrainer):

    def __init__(
        self,
        net: torch.nn.Module,
        params: utils.Hyperparams,
        metadata: datasets.Metadata,
        state: nnts.trainers.TrainerState = None,
        model_path: str = "best_model.pt",
    ):
        super().__init__(state, params)
        self.net = net
        self.metadata = metadata
        utils.makedirs_if_not_exists(params.model_file_path)
        self.path = os.path.join(params.model_file_path, model_path)
        self.loss_fn = params.loss_fn
        self.Optimizer = params.optimizer

    def before_train(self, train_dl):
        print(self.net)
        self.optimizer = self.Optimizer(
            self.net.parameters(),
            lr=self.params.lr,
            weight_decay=self.params.weight_decay,
        )
        if self.params.scheduler == utils.Scheduler.REDUCE_LR_ON_PLATEAU:
            self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer, mode="min", factor=0.5, patience=self.params.patience
            )
        elif self.params.scheduler == utils.Scheduler.STEP_LR:
            self.scheduler = torch.optim.lr_scheduler.StepLR(
                self.optimizer, step_size=1, gamma=0.5
            )
        elif self.params.scheduler == utils.Scheduler.ONE_CYCLE:
            self.scheduler = torch.optim.lr_scheduler.OneCycleLR(
                self.optimizer,
                max_lr=self.params.lr * 3,
                steps_per_epoch=self.params.batches_per_epoch,
                epochs=self.params.epochs + 2,
            )
        self.best_loss = np.inf

    def before_train_epoch(self) -> None:
        self.net.train()

    def after_train_epoch(self) -> None:
        if self.state.train_loss < self.state.best_loss:
            print(f"saving model to {self.path}")
            torch.save(self.net.state_dict(), self.path)
            self.state.best_loss = self.state.train_loss

        if self.params.scheduler == utils.Scheduler.REDUCE_LR_ON_PLATEAU:
            self.scheduler.step(self.state.train_loss)
        elif self.params.scheduler == utils.Scheduler.STEP_LR:
            self.scheduler.step()

        print(f"Epoch {self.state.epoch} Train Loss: {self.state.train_loss}")

    def _train_batch(self, i, batch):
        self.optimizer.zero_grad()
        y_hat, y = self.net.train_output(
            batch,
            self.metadata.prediction_length,
            self.metadata.context_length,
            training_method=self.params.training_method,
        )

        L = self.loss_fn(y_hat, y)
        L.backward()
        # torch.nn.utils.clip_grad_value_(self.net.parameters(), 10.0)
        self.optimizer.step()

        if self.params.scheduler == utils.Scheduler.ONE_CYCLE:
            self.scheduler.step()  # This is required for OneCycleLR
        return L

    def _validate_batch(self, i, batch):
        pass

    def create_evaluator(self) -> nnts.trainers.Evaluator:
        self.events.notify(nnts.trainers.EpochBestModel(self.path))
        state_dict = torch.load(self.path)
        self.net.load_state_dict(state_dict)
        print(f"Best model loaded from {self.path}")
        return TorchEvaluator(self.net)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from math import sqrt
from typing import Any, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor, nn

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    freq: str
    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.1
    batch_size: int = 32
    lr: float = 0.005
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path: str = f"logs"
    d_model: int = 512
    embed: str = "timeF"
    use_norm: bool = True
    class_strategy: str = "projection"
    factor: int = 1
    n_heads: int = 8
    output_attention: bool = True
    d_ff: int = 512
    activation: str = "gelu"
    e_layers: int = 2
    enc_in: int = 1
    dec_in: int = 1
    c_out: int = 1


def get_mutlivariate_params():
    params = Hyperparams(
        freq="h",
        optimizer=torch.optim.Adam,
        loss_fn=torch.nn.MSELoss(),
        batch_size=32,
        batches_per_epoch=250,
        training_method=TrainingMethod.DMS,
        model_file_path="logs",
        epochs=10,
        scheduler=Scheduler.STEP_LR,
        lr=0.0001,
        weight_decay=0.0,
        enc_in=7,
        dec_in=7,
        c_out=7,
        d_model=128,
        d_ff=128,
    )
    return params


class TriangularCausalMask:
    def __init__(self, B, L, device="cpu"):
        mask_shape = [B, 1, L, L]
        with torch.no_grad():
            self._mask = torch.triu(
                torch.ones(mask_shape, dtype=torch.bool), diagonal=1
            ).to(device)

    @property
    def mask(self):
        return self._mask


class FullAttention(nn.Module):
    def __init__(
        self,
        mask_flag=True,
        factor=5,
        scale=None,
        attention_dropout=0.1,
        output_attention=False,
    ):
        super(FullAttention, self).__init__()
        self.scale = scale
        self.mask_flag = mask_flag
        self.output_attention = output_attention
        self.dropout = nn.Dropout(attention_dropout)

    def forward(self, queries, keys, values, attn_mask, tau=None, delta=None):
        B, L, H, E = queries.shape
        _, S, _, D = values.shape
        scale = self.scale or 1.0 / sqrt(E)

        scores = torch.einsum("blhe,bshe->bhls", queries, keys)

        if self.mask_flag:
            if attn_mask is None:
                attn_mask = TriangularCausalMask(B, L, device=queries.device)

            scores.masked_fill_(attn_mask.mask, -np.inf)

        A = self.dropout(torch.softmax(scale * scores, dim=-1))
        V = torch.einsum("bhls,bshd->blhd", A, values)

        if self.output_attention:
            return (V.contiguous(), A)
        else:
            return (V.contiguous(), None)


class AttentionLayer(nn.Module):
    def __init__(self, attention, d_model, n_heads, d_keys=None, d_values=None):
        super(AttentionLayer, self).__init__()

        d_keys = d_keys or (d_model // n_heads)
        d_values = d_values or (d_model // n_heads)

        self.inner_attention = attention
        self.query_projection = nn.Linear(d_model, d_keys * n_heads)
        self.key_projection = nn.Linear(d_model, d_keys * n_heads)
        self.value_projection = nn.Linear(d_model, d_values * n_heads)
        self.out_projection = nn.Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads

    def forward(self, queries, keys, values, attn_mask, tau=None, delta=None):
        B, L, _ = queries.shape
        _, S, _ = keys.shape
        H = self.n_heads

        queries = self.query_projection(queries).view(B, L, H, -1)
        keys = self.key_projection(keys).view(B, S, H, -1)
        values = self.value_projection(values).view(B, S, H, -1)

        out, attn = self.inner_attention(
            queries, keys, values, attn_mask, tau=tau, delta=delta
        )
        out = out.view(B, L, -1)

        return self.out_projection(out), attn


class DataEmbedding_inverted(nn.Module):
    def __init__(self, c_in, d_model, embed_type="fixed", freq="h", dropout=0.1):
        super(DataEmbedding_inverted, self).__init__()
        self.value_embedding = nn.Linear(c_in, d_model)
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x, x_mark):
        x = x.permute(0, 2, 1)
        # x: [Batch Variate Time]
        if x_mark is None:
            x = self.value_embedding(x)
        else:
            # the potential to take covariates (e.g. timestamps) as tokens
            x = self.value_embedding(torch.cat([x, x_mark.permute(0, 2, 1)], 1))
        # x: [Batch Variate d_model]
        return self.dropout(x)


class EncoderLayer(nn.Module):
    def __init__(self, attention, d_model, d_ff=None, dropout=0.1, activation="relu"):
        super(EncoderLayer, self).__init__()
        d_ff = d_ff or 4 * d_model
        self.attention = attention
        self.conv1 = nn.Conv1d(in_channels=d_model, out_channels=d_ff, kernel_size=1)
        self.conv2 = nn.Conv1d(in_channels=d_ff, out_channels=d_model, kernel_size=1)
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, attn_mask=None, tau=None, delta=None):
        new_x, attn = self.attention(x, x, x, attn_mask=attn_mask, tau=tau, delta=delta)
        x = x + self.dropout(new_x)

        y = x = self.norm1(x)
        y = self.dropout(self.activation(self.conv1(y.transpose(-1, 1))))
        y = self.dropout(self.conv2(y).transpose(-1, 1))

        return self.norm2(x + y), attn


class Encoder(nn.Module):
    def __init__(self, attn_layers, conv_layers=None, norm_layer=None):
        super(Encoder, self).__init__()
        self.attn_layers = nn.ModuleList(attn_layers)
        self.conv_layers = (
            nn.ModuleList(conv_layers) if conv_layers is not None else None
        )
        self.norm = norm_layer

    def forward(self, x, attn_mask=None, tau=None, delta=None):
        # x [B, L, D]
        attns = []
        if self.conv_layers is not None:
            for i, (attn_layer, conv_layer) in enumerate(
                zip(self.attn_layers, self.conv_layers)
            ):
                delta = delta if i == 0 else None
                x, attn = attn_layer(x, attn_mask=attn_mask, tau=tau, delta=delta)
                x = conv_layer(x)
                attns.append(attn)
            x, attn = self.attn_layers[-1](x, tau=tau, delta=None)
            attns.append(attn)
        else:
            for attn_layer in self.attn_layers:
                x, attn = attn_layer(x, attn_mask=attn_mask, tau=tau, delta=delta)
                attns.append(attn)

        if self.norm is not None:
            x = self.norm(x)

        return x, attns


class iTransformer(nn.Module):
    """
    Paper link: https://arxiv.org/abs/2310.06625
    """

    def __init__(self, context_length, prediction_length, configs: Hyperparams):
        super(iTransformer, self).__init__()
        self.seq_len = context_length
        self.pred_len = prediction_length
        self.output_attention = configs.output_attention
        self.use_norm = configs.use_norm
        # Embedding
        self.enc_embedding = DataEmbedding_inverted(
            context_length,
            configs.d_model,
            configs.embed,
            configs.freq,
            configs.dropout,
        )
        self.class_strategy = configs.class_strategy
        # Encoder-only architecture
        self.encoder = Encoder(
            [
                EncoderLayer(
                    AttentionLayer(
                        FullAttention(
                            False,
                            configs.factor,
                            attention_dropout=configs.dropout,
                            output_attention=configs.output_attention,
                        ),
                        configs.d_model,
                        configs.n_heads,
                    ),
                    configs.d_model,
                    configs.d_ff,
                    dropout=configs.dropout,
                    activation=configs.activation,
                )
                for l in range(configs.e_layers)
            ],
            norm_layer=torch.nn.LayerNorm(configs.d_model),
        )
        self.projector = nn.Linear(configs.d_model, prediction_length, bias=True)

    def forecast(self, x_enc, x_mark_enc):
        if self.use_norm:
            # Normalization from Non-stationary Transformer
            means = x_enc.mean(1, keepdim=True).detach()
            x_enc = x_enc - means
            stdev = torch.sqrt(
                torch.var(x_enc, dim=1, keepdim=True, unbiased=False) + 1e-5
            )
            x_enc /= stdev

        _, _, N = x_enc.shape  # B L N
        # B: batch_size;    E: d_model;
        # L: seq_len;       S: pred_len;
        # N: number of variate (tokens), can also includes covariates

        # Embedding
        # B L N -> B N E                (B L N -> B L E in the vanilla Transformer)
        enc_out = self.enc_embedding(
            x_enc, x_mark_enc
        )  # covariates (e.g timestamp) can be also embedded as tokens

        # B N E -> B N E                (B L E -> B L E in the vanilla Transformer)
        # the dimensions of embedded time series has been inverted, and then processed by native attn, layernorm and ffn modules
        enc_out, attns = self.encoder(enc_out, attn_mask=None)

        # B N E -> B N S -> B S N
        dec_out = self.projector(enc_out).permute(0, 2, 1)[
            :, :, :N
        ]  # filter the covariates

        if self.use_norm:
            # De-Normalization from Non-stationary Transformer
            dec_out = dec_out * (
                stdev[:, 0, :].unsqueeze(1).repeat(1, self.pred_len, 1)
            )
            dec_out = dec_out + (
                means[:, 0, :].unsqueeze(1).repeat(1, self.pred_len, 1)
            )

        return dec_out

    def forward(self, x_enc, x_mark_enc):
        dec_out = self.forecast(x_enc, x_mark_enc)
        return dec_out[:, -self.pred_len :, :]  # [B, L, D]

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        pad_mask = batch.pad_mask[:, : self.seq_len]
        y_hat = self(x, None)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x, None)
        return y_hat

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        pad_mask = batch.pad_mask[:, : self.seq_len]
        y_hat = self(x, None)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x, None)
        return y_hat



from dataclasses import dataclass
from typing import Any, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.0
    batch_size: int = 32
    lr: float = 0.005
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path: str = f"logs"
    individual: bool = True
    kernel_size = 25
    scaling_fn: Any = None
    enc_in: int = 1


def get_mutlivariate_params():
    params = Hyperparams(
        optimizer=torch.optim.Adam,
        loss_fn=torch.nn.MSELoss(),
        batch_size=32,
        batches_per_epoch=250,
        training_method=TrainingMethod.DMS,
        model_file_path="logs",
        epochs=10,
        scheduler=Scheduler.STEP_LR,
        lr=0.005,
        weight_decay=0.0,
    )
    return params


class moving_avg(nn.Module):
    """
    Moving average block to highlight the trend of time series
    """

    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class series_decomp(nn.Module):
    """
    Series decomposition block
    """

    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


class DLinear(nn.Module):
    """
    DLinear
    """

    def __init__(self, h: int, input_size: int, c_in: int, configs: Hyperparams):
        super(DLinear, self).__init__()
        print("enc_in", configs.enc_in)
        self.seq_len = input_size
        self.pred_len = h
        self.scaling_fn = configs.scaling_fn

        # Decompsition Kernel Size
        self.decompsition = series_decomp(configs.kernel_size)
        self.individual = configs.individual
        self.channels = c_in

        if self.individual:
            self.Linear_Seasonal = nn.ModuleList()
            self.Linear_Trend = nn.ModuleList()

            for i in range(self.channels):
                self.Linear_Seasonal.append(nn.Linear(self.seq_len, self.pred_len))
                self.Linear_Trend.append(nn.Linear(self.seq_len, self.pred_len))

                # Use this two lines if you want to visualize the weights
                # self.Linear_Seasonal[i].weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
                # self.Linear_Trend[i].weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
        else:
            self.Linear_Seasonal = nn.Linear(self.seq_len, self.pred_len)
            self.Linear_Trend = nn.Linear(self.seq_len, self.pred_len)

            # Use this two lines if you want to visualize the weights
            # self.Linear_Seasonal.weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
            # self.Linear_Trend.weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))

    def forward(self, x):
        # x: [Batch, Input length, Channel]
        if self.scaling_fn is not None:
            target_scale = self.scaling_fn(x)
            x = x / target_scale

        seasonal_init, trend_init = self.decompsition(x)
        seasonal_init, trend_init = seasonal_init.permute(0, 2, 1), trend_init.permute(
            0, 2, 1
        )
        if self.individual:
            seasonal_output = torch.zeros(
                [seasonal_init.size(0), seasonal_init.size(1), self.pred_len],
                dtype=seasonal_init.dtype,
            ).to(seasonal_init.device)
            trend_output = torch.zeros(
                [trend_init.size(0), trend_init.size(1), self.pred_len],
                dtype=trend_init.dtype,
            ).to(trend_init.device)
            for i in range(self.channels):
                seasonal_output[:, i, :] = self.Linear_Seasonal[i](
                    seasonal_init[:, i, :]
                )
                trend_output[:, i, :] = self.Linear_Trend[i](trend_init[:, i, :])
        else:
            seasonal_output = self.Linear_Seasonal(seasonal_init)
            trend_output = self.Linear_Trend(trend_init)

        x = seasonal_output + trend_output
        x = x.permute(0, 2, 1)  # to [Batch, Output length, Channel]

        if self.scaling_fn is not None:
            x = x * target_scale
        return x

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        y_hat = self(x)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x)
        return y_hat



from typing import Any, List, Tuple

import matplotlib.pyplot as plt
import numpy as np


def subplot_forecast_horizon(
    selected_scenarios: List[Any],
    dataset: str,
    model_name: str,
    ax: plt.Axes,
    loader_fn: callable,
):
    """plot forecast horizon for a given model and dataset"""
    for scenario in selected_scenarios:
        forecast_horizon_metrics = loader_fn(f"{model_name}/{dataset}/{scenario.name}")
        ax.plot(forecast_horizon_metrics, label=scenario.name)

    ax.set_xlim(1, len(forecast_horizon_metrics))
    ax.set_xlabel("Forecast Horizon")
    ax.set_ylabel("Error (sMAPE)")
    ax.set_title(f"{model_name}")
    ax.legend()


def plot_forecast_horizon_trajectories(
    selected_scenarios: List[Any],
    dataset: str,
    covariates: List[int] | int,
    model_names: List[str] = None,
    loader_fn: callable = None,
    figsize: Tuple[int, int] = (20, 5),
    path: str = None,
):
    # FORECAST HORIZON TRAJECTORIES
    if model_names is None:
        raise ValueError("Please provide a list of model names to compare")
    if loader_fn is None:
        raise ValueError("Please provide a loader function to load the data")

    fig, axes = plt.subplots(
        nrows=1, ncols=len(model_names), figsize=figsize, sharey=True
    )

    for i, model_name in enumerate(model_names):
        ax = axes[i] if len(model_names) > 1 else axes
        subplot_forecast_horizon(
            selected_scenarios, dataset, model_name, ax, loader_fn=loader_fn
        )

    plt.suptitle(f"{dataset} forecast horizon trajectories {covariates} covariates")
    plt.tight_layout()
    if path:
        if isinstance(covariates, list):
            covariates = "_".join(map(str, covariates))
        model_name = "_".join(model_names)
        full_path = f"{path}/{model_name}_{dataset}_k_{covariates}_trajectory.png"
        print("Saving to", full_path)
        plt.savefig(full_path)
    else:
        plt.show()

    return plt


def plot(df_test, prediction_length, start_idx=0):
    num_plots = min(len(df_test), 4)
    fig, axes = plt.subplots(
        nrows=num_plots // 2 + num_plots % 2, ncols=min(num_plots, 2), figsize=(20, 10)
    )
    axes = np.ravel(axes)  # Flatten the axes array

    for idx, ax in enumerate(axes):
        if idx < len(df_test):
            df_test[start_idx + idx].set_index("ds").tail(prediction_length * 5)[
                ["y", "y_hat"]
            ].plot(ax=ax)
        else:
            ax.axis("off")  # Hide empty subplots if df_test length is less than 4
    return fig


def plot_forecasts_vs_actuals(y, y_hat):
    """
    Plots the forecasts vs actuals for the first 4 time series.

    Parameters:
    - y: Actual values tensor of shape [320, 24]
    - y_hat: Forecasted values tensor of shape [320, 24]
    """
    num_series_to_plot = 4
    timesteps = y.shape[1]

    fig, axes = plt.subplots(num_series_to_plot, 1, figsize=(10, 10), sharex=True)
    fig.suptitle("Forecasts vs Actuals for First 4 Time Series")

    for i in range(num_series_to_plot):
        axes[i].plot(range(timesteps), y[i], label="Actual", color="blue")
        axes[i].plot(
            range(timesteps),
            y_hat[i],
            label="Forecast",
            color="orange",
            linestyle="dashed",
        )
        axes[i].set_title(f"Time Series {i+1}")
        axes[i].legend()

    plt.xlabel("Timesteps")
    plt.tight_layout(
        rect=[0, 0, 1, 0.96]
    )  # Adjust layout to make room for the main title
    plt.show()


import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots


# Function to plot the forecasts vs actuals for the first 4 time series
def plotly_forecasts_vs_actuals(y, y_hat):
    """
    Plots the forecasts vs actuals for the first 4 time series using Plotly.

    Parameters:
    - y: Actual values tensor of shape [320, 24]
    - y_hat: Forecasted values tensor of shape [320, 24]
    """
    y_hat = y_hat.squeeze(-1)
    y = y.squeeze(-1)
    num_series_to_plot = 4
    timesteps = y.shape[1]

    fig = make_subplots(
        rows=num_series_to_plot,
        cols=1,
        shared_xaxes=True,
        subplot_titles=[f"Time Series {i+1}" for i in range(num_series_to_plot)],
    )

    for i in range(num_series_to_plot):
        fig.add_trace(
            go.Scatter(
                x=list(range(timesteps)),
                y=y[i],
                mode="lines",
                name=f"Actual {i+1}",
                line=dict(color="blue"),
            ),
            row=i + 1,
            col=1,
        )
        fig.add_trace(
            go.Scatter(
                x=list(range(timesteps)),
                y=y_hat[i],
                mode="lines",
                name=f"Forecast {i+1}",
                line=dict(color="orange", dash="dash"),
            ),
            row=i + 1,
            col=1,
        )

    fig.update_layout(
        height=1000,
        width=800,
        title_text="Forecasts vs Actuals for First 4 Time Series",
        showlegend=True,
    )

    fig.update_xaxes(title_text="Timesteps", row=num_series_to_plot, col=1)

    fig.show()



from dataclasses import dataclass
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.3
    batch_size: int = 32
    lr: float = 1e-3
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path = f"logs"

    hidden_size: int = 256
    num_encoder_layers: int = 2
    num_decoder_layers: int = 2
    decoder_output_dim: int = 8
    temporal_decoder_dim: int = 128
    output_dim: int = 1


class ResidualBlock(nn.Module):
    """Residual Block. We use the residual block as the basic layer in our architecture. It is an MLP with one
    hidden layer with ReLU activation. It also has a skip connection that is fully linear. We use dropout on the
    linear layer that maps the hidden layer to the output and also use layer norm at the output.
    We separate the model into encoding and decoding sections. The encoding section has a novel feature
    projection step followed by a dense MLP encoder. The decoder section consists of a dense decoder followed
    by a novel temporal decoder. Note that the dense encoder (green block with ne layers) and decoder blocks
    (yellow block with nd layers) in Figure 1 can be merged into a single block. For the sake of exposition we keep
    them separate as we tune the hidden layer size in the two blocks separately. Also the last layer of the decoder
    block is unique in the sense that its output dimension needs to be H × p before the reshape operation.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_size: int,
        output_dim: int,
        dropout: float = 0.0,
        norm: bool = True,
    ):
        super(ResidualBlock, self).__init__()
        self.hidden_size = hidden_size
        self.dropout = dropout

        self.fc1 = nn.Linear(input_dim, hidden_size)
        self.fc2 = nn.Linear(hidden_size, output_dim)
        self.skip = nn.Linear(input_dim, output_dim)
        self.dropout = nn.Dropout(dropout)
        if norm:
            self.layer_norm = nn.LayerNorm(output_dim)
        self.norm = norm

    def forward(self, x):
        out = self.fc2(F.relu(self.fc1(x)))
        out = self.dropout(out)
        skip = self.skip(x)
        out += skip
        if self.norm:
            out = self.layer_norm(out)
        return out


class TiDE(nn.Module):

    def __init__(
        self,
        h: int,
        input_size: int,
        configs: Hyperparams,
    ):
        super(TiDE, self).__init__()
        self.h = h
        self.seq_len = input_size

        dense_encoder_layers = [
            ResidualBlock(
                input_dim=input_size if i == 0 else configs.hidden_size,
                hidden_size=configs.hidden_size,
                output_dim=configs.hidden_size,
                dropout=configs.dropout,
            )
            for i in range(configs.num_encoder_layers)
        ]
        self.dense_encoder = nn.Sequential(*dense_encoder_layers)

        # Decoder
        decoder_output_size = configs.decoder_output_dim * h
        dense_decoder_layers = [
            ResidualBlock(
                input_dim=configs.hidden_size,
                hidden_size=configs.hidden_size,
                output_dim=(
                    decoder_output_size
                    if i == configs.num_decoder_layers - 1
                    else configs.hidden_size
                ),
                dropout=configs.dropout,
            )
            for i in range(configs.num_decoder_layers)
        ]
        self.dense_decoder = nn.Sequential(*dense_decoder_layers)

        self.temporal_decoder = ResidualBlock(
            input_dim=configs.decoder_output_dim,
            hidden_size=configs.temporal_decoder_dim,
            output_dim=configs.output_dim,
            dropout=configs.dropout,
            norm=True,
        )

        self.global_skip = nn.Linear(
            in_features=input_size, out_features=h * configs.output_dim
        )

    def forward(self, x: torch.Tensor, pad_mask: torch.Tensor) -> torch.Tensor:
        batch_size, seq_len = x.shape[:2]

        # Flatten insample_y
        x = x.reshape(batch_size, -1)  #   [B, L, 1] -> [B, L]

        # Global skip connection
        x_skip = self.global_skip(x)  #   [B, L] -> [B, h * n_outputs]
        x_skip = x_skip.reshape(
            batch_size, self.h, -1
        )  #   [B, h * n_outputs] -> [B, h, n_outputs]

        # Dense encoder
        x = self.dense_encoder(
            x
        )  #   [B, L * (1 + 2 * temporal_width) + h * temporal_width + S] -> [B, hidden_size]

        # Dense decoder
        x = self.dense_decoder(x)  #   [B, hidden_size] ->  [B, decoder_output_dim * h]
        x = x.reshape(
            batch_size, self.h, -1
        )  #   [B, decoder_output_dim * h] -> [B, h, decoder_output_dim]

        # Temporal decoder
        x = self.temporal_decoder(
            x
        )  #  [B, h, temporal_width + decoder_output_dim] -> [B, h, n_outputs]

        # Map to output domain
        forecast = x + x_skip
        return forecast

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        x_pad_mask = batch.pad_mask[:, : self.seq_len]
        y_hat = self(x, x_pad_mask)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x, pad_mask)
        return y_hat



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor, nn

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.1
    batch_size: int = 32
    lr: float = 0.01
    epochs: int = 100
    patience: int = 20
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path: str = f"logs"
    revin: bool = True
    affine: bool = True
    subtract_last: bool = False
    task_name: str = "short_term_forecast"
    label_len: int = 0
    down_sampling_window: int = 2
    down_sampling_layers: int = 1
    d_model: int = 32
    d_ff: int = 32
    e_layers: int = 4
    factor: int = 3
    enc_in: int = 1
    dec_in: int = 1
    c_out: int = 1
    down_sampling_method: str = "avg"
    channel_independence: int = 1
    moving_avg: int = 25
    embed: str = "timeF"
    freq: str = None  # h, d, w, m
    use_future_temporal_feature: int = 0
    decomp_method: str = "moving_avg"
    use_norm: int = 1
    num_class: int = None


def get_mutlivariate_params():
    params = Hyperparams(
        optimizer=torch.optim.Adam,
        loss_fn=torch.nn.MSELoss(),
        batch_size=32,
        batches_per_epoch=250,
        training_method=TrainingMethod.DMS,
        model_file_path="logs",
        epochs=10,
        scheduler=Scheduler.STEP_LR,
        lr=0.001,
        weight_decay=0.0,
        blocks=4,
        dropout=0.9,
        num_blocks=4,
        ff_dim=256,
        output_channels=7,
    )
    return params


class PositionalEmbedding(nn.Module):
    def __init__(self, d_model, max_len=5000):
        super(PositionalEmbedding, self).__init__()
        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model).float()
        pe.require_grad = False

        position = torch.arange(0, max_len).float().unsqueeze(1)
        div_term = (
            torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)
        ).exp()

        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        return self.pe[:, : x.size(1)]


class TokenEmbedding(nn.Module):
    def __init__(self, c_in, d_model):
        super(TokenEmbedding, self).__init__()
        padding = 1 if torch.__version__ >= "1.5.0" else 2
        self.tokenConv = nn.Conv1d(
            in_channels=c_in,
            out_channels=d_model,
            kernel_size=3,
            padding=padding,
            padding_mode="circular",
            bias=False,
        )
        for m in self.modules():
            if isinstance(m, nn.Conv1d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_in", nonlinearity="leaky_relu"
                )

    def forward(self, x):
        x = self.tokenConv(x.permute(0, 2, 1)).transpose(1, 2)
        return x


class FixedEmbedding(nn.Module):
    def __init__(self, c_in, d_model):
        super(FixedEmbedding, self).__init__()

        w = torch.zeros(c_in, d_model).float()
        w.require_grad = False

        position = torch.arange(0, c_in).float().unsqueeze(1)
        div_term = (
            torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)
        ).exp()

        w[:, 0::2] = torch.sin(position * div_term)
        w[:, 1::2] = torch.cos(position * div_term)

        self.emb = nn.Embedding(c_in, d_model)
        self.emb.weight = nn.Parameter(w, requires_grad=False)

    def forward(self, x):
        return self.emb(x).detach()


class TemporalEmbedding(nn.Module):
    def __init__(self, d_model, embed_type="fixed", freq="h"):
        super(TemporalEmbedding, self).__init__()

        minute_size = 4
        hour_size = 24
        weekday_size = 7
        day_size = 32
        month_size = 13

        Embed = FixedEmbedding if embed_type == "fixed" else nn.Embedding
        if freq == "t":
            self.minute_embed = Embed(minute_size, d_model)
        self.hour_embed = Embed(hour_size, d_model)
        self.weekday_embed = Embed(weekday_size, d_model)
        self.day_embed = Embed(day_size, d_model)
        self.month_embed = Embed(month_size, d_model)

    def forward(self, x):
        x = x.long()
        minute_x = (
            self.minute_embed(x[:, :, 4]) if hasattr(self, "minute_embed") else 0.0
        )
        hour_x = self.hour_embed(x[:, :, 3])
        weekday_x = self.weekday_embed(x[:, :, 2])
        day_x = self.day_embed(x[:, :, 1])
        month_x = self.month_embed(x[:, :, 0])

        return hour_x + weekday_x + day_x + month_x + minute_x


class TimeFeatureEmbedding(nn.Module):
    def __init__(self, d_model, embed_type="timeF", freq="h"):
        super(TimeFeatureEmbedding, self).__init__()

        # freq_map = {"h": 4, "t": 5, "s": 6, "m": 1, "a": 1, "w": 2, "d": 3, "b": 3}
        freq_map = {
            "ME": 1,
            "QE": 1,
            "YE": 1,
            "1D": 1,
            "1H": 4,
            "1W": 2,
            "30min": 5,
            "10min": 5,
            "h": 4,
        }
        d_inp = freq_map[freq]
        self.embed = nn.Linear(d_inp, d_model, bias=False)

    def forward(self, x):
        return self.embed(x)


class DataEmbedding_wo_pos(nn.Module):
    def __init__(self, c_in, d_model, embed_type="fixed", freq="h", dropout=0.1):
        super(DataEmbedding_wo_pos, self).__init__()

        self.value_embedding = TokenEmbedding(c_in=c_in, d_model=d_model)
        self.position_embedding = PositionalEmbedding(d_model=d_model)
        self.temporal_embedding = (
            TemporalEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
            if embed_type != "timeF"
            else TimeFeatureEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x, x_mark):
        if x is None and x_mark is not None:
            return self.temporal_embedding(x_mark)
        if x_mark is None:
            x = self.value_embedding(x)
        else:
            x = self.value_embedding(x) + self.temporal_embedding(x_mark)
        return self.dropout(x)


class moving_avg(nn.Module):
    """
    Moving average block to highlight the trend of time series
    """

    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class series_decomp(nn.Module):
    """
    Series decomposition block
    """

    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


class RevIN(nn.Module):
    def __init__(self, num_features: int, eps=1e-5, affine=True, subtract_last=False):
        """
        :param num_features: the number of features or channels
        :param eps: a value added for numerical stability
        :param affine: if True, RevIN has learnable affine parameters
        """
        super(RevIN, self).__init__()
        self.num_features = num_features
        self.eps = eps
        self.affine = affine
        self.subtract_last = subtract_last
        if self.affine:
            self._init_params()

    def forward(self, x, mode: str):
        if mode == "norm":
            self._get_statistics(x)
            x = self._normalize(x)
        elif mode == "denorm":
            x = self._denormalize(x)
        else:
            raise NotImplementedError
        return x

    def _init_params(self):
        # initialize RevIN params: (C,)
        self.affine_weight = nn.Parameter(torch.ones(self.num_features))
        self.affine_bias = nn.Parameter(torch.zeros(self.num_features))

    def _get_statistics(self, x):
        dim2reduce = tuple(range(1, x.ndim - 1))
        if self.subtract_last:
            self.last = x[:, -1, :].unsqueeze(1)
        else:
            self.mean = torch.mean(x, dim=dim2reduce, keepdim=True).detach()
        self.stdev = torch.sqrt(
            torch.var(x, dim=dim2reduce, keepdim=True, unbiased=False) + self.eps
        ).detach()

    def _normalize(self, x):
        if self.subtract_last:
            x = x - self.last
        else:
            x = x - self.mean
        x = x / self.stdev
        if self.affine:
            x = x * self.affine_weight
            x = x + self.affine_bias
        return x

    def _denormalize(self, x):
        if self.affine:
            x = x - self.affine_bias
            x = x / (self.affine_weight + self.eps * self.eps)
        x = x * self.stdev
        if self.subtract_last:
            x = x + self.last
        else:
            x = x + self.mean
        return x


class DFT_series_decomp(nn.Module):
    """
    Series decomposition block
    """

    def __init__(self, top_k=5):
        super(DFT_series_decomp, self).__init__()
        self.top_k = top_k

    def forward(self, x):
        xf = torch.fft.rfft(x)
        freq = abs(xf)
        freq[0] = 0
        top_k_freq, top_list = torch.topk(freq, self.top_k)
        xf[freq <= top_k_freq.min()] = 0
        x_season = torch.fft.irfft(xf)
        x_trend = x - x_season
        return x_season, x_trend


class MultiScaleSeasonMixing(nn.Module):
    """
    Bottom-up mixing season pattern
    """

    def __init__(self, seq_len: int, pred_len: int, params: Hyperparams):
        super(MultiScaleSeasonMixing, self).__init__()

        self.down_sampling_layers = torch.nn.ModuleList(
            [
                nn.Sequential(
                    torch.nn.Linear(
                        seq_len // (params.down_sampling_window**i),
                        seq_len // (params.down_sampling_window ** (i + 1)),
                    ),
                    nn.GELU(),
                    torch.nn.Linear(
                        seq_len // (params.down_sampling_window ** (i + 1)),
                        seq_len // (params.down_sampling_window ** (i + 1)),
                    ),
                )
                for i in range(params.down_sampling_layers)
            ]
        )

    def forward(self, season_list):

        # mixing high->low
        out_high = season_list[0]
        out_low = season_list[1]
        out_season_list = [out_high.permute(0, 2, 1)]

        for i in range(len(season_list) - 1):
            out_low_res = self.down_sampling_layers[i](out_high)
            out_low = out_low + out_low_res
            out_high = out_low
            if i + 2 <= len(season_list) - 1:
                out_low = season_list[i + 2]
            out_season_list.append(out_high.permute(0, 2, 1))

        return out_season_list


class MultiScaleTrendMixing(nn.Module):
    """
    Top-down mixing trend pattern
    """

    def __init__(self, seq_len: int, pred_len: int, params: Hyperparams):
        super(MultiScaleTrendMixing, self).__init__()

        self.up_sampling_layers = torch.nn.ModuleList(
            [
                nn.Sequential(
                    torch.nn.Linear(
                        seq_len // (params.down_sampling_window ** (i + 1)),
                        seq_len // (params.down_sampling_window**i),
                    ),
                    nn.GELU(),
                    torch.nn.Linear(
                        seq_len // (params.down_sampling_window**i),
                        seq_len // (params.down_sampling_window**i),
                    ),
                )
                for i in reversed(range(params.down_sampling_layers))
            ]
        )

    def forward(self, trend_list):

        # mixing low->high
        trend_list_reverse = trend_list.copy()
        trend_list_reverse.reverse()
        out_low = trend_list_reverse[0]
        out_high = trend_list_reverse[1]
        out_trend_list = [out_low.permute(0, 2, 1)]

        for i in range(len(trend_list_reverse) - 1):
            out_high_res = self.up_sampling_layers[i](out_low)
            out_high = out_high + out_high_res
            out_low = out_high
            if i + 2 <= len(trend_list_reverse) - 1:
                out_high = trend_list_reverse[i + 2]
            out_trend_list.append(out_low.permute(0, 2, 1))

        out_trend_list.reverse()
        return out_trend_list


class PastDecomposableMixing(nn.Module):
    def __init__(self, seq_len: int, pred_len: int, params: Hyperparams):
        super(PastDecomposableMixing, self).__init__()
        self.seq_len = seq_len
        self.pred_len = pred_len
        self.down_sampling_window = params.down_sampling_window

        self.layer_norm = nn.LayerNorm(params.d_model)
        self.dropout = nn.Dropout(params.dropout)
        self.channel_independence = params.channel_independence

        if params.decomp_method == "moving_avg":
            self.decompsition = series_decomp(params.moving_avg)
        elif params.decomp_method == "dft_decomp":
            self.decompsition = DFT_series_decomp(params.top_k)
        else:
            raise ValueError("decompsition is error")

        if params.channel_independence == 0:
            self.cross_layer = nn.Sequential(
                nn.Linear(in_features=params.d_model, out_features=params.d_ff),
                nn.GELU(),
                nn.Linear(in_features=params.d_ff, out_features=params.d_model),
            )

        # Mixing season
        self.mixing_multi_scale_season = MultiScaleSeasonMixing(
            seq_len, pred_len, params
        )

        # Mxing trend
        self.mixing_multi_scale_trend = MultiScaleTrendMixing(seq_len, pred_len, params)

        self.out_cross_layer = nn.Sequential(
            nn.Linear(in_features=params.d_model, out_features=params.d_ff),
            nn.GELU(),
            nn.Linear(in_features=params.d_ff, out_features=params.d_model),
        )

    def forward(self, x_list):
        length_list = []
        for x in x_list:
            _, T, _ = x.size()
            length_list.append(T)

        # Decompose to obtain the season and trend
        season_list = []
        trend_list = []
        for x in x_list:
            season, trend = self.decompsition(x)
            if self.channel_independence == 0:
                season = self.cross_layer(season)
                trend = self.cross_layer(trend)
            season_list.append(season.permute(0, 2, 1))
            trend_list.append(trend.permute(0, 2, 1))

        # bottom-up season mixing
        out_season_list = self.mixing_multi_scale_season(season_list)
        # top-down trend mixing
        out_trend_list = self.mixing_multi_scale_trend(trend_list)

        out_list = []
        for ori, out_season, out_trend, length in zip(
            x_list, out_season_list, out_trend_list, length_list
        ):
            out = out_season + out_trend
            if self.channel_independence:
                out = ori + self.out_cross_layer(out)
            out_list.append(out[:, :length, :])
        return out_list


class TimeMixer(nn.Module):

    def __init__(
        self,
        sequence_length,
        prediction_length,
        params: Hyperparams,
    ):
        super(TimeMixer, self).__init__()
        self.params = params
        self.task_name = params.task_name
        self.seq_len = sequence_length
        self.label_len = params.label_len
        self.pred_len = prediction_length
        self.down_sampling_window = params.down_sampling_window
        self.channel_independence = params.channel_independence
        self.pdm_blocks = nn.ModuleList(
            [
                PastDecomposableMixing(self.seq_len, self.pred_len, params)
                for _ in range(params.e_layers)
            ]
        )

        self.preprocess = series_decomp(params.moving_avg)
        self.enc_in = params.enc_in
        self.use_future_temporal_feature = params.use_future_temporal_feature

        if self.channel_independence == 1:
            self.enc_embedding = DataEmbedding_wo_pos(
                1, params.d_model, params.embed, params.freq, params.dropout
            )
        else:
            self.enc_embedding = DataEmbedding_wo_pos(
                params.enc_in,
                params.d_model,
                params.embed,
                params.freq,
                params.dropout,
            )

        self.layer = params.e_layers

        self.normalize_layers = torch.nn.ModuleList(
            [
                RevIN(
                    self.params.enc_in,
                    affine=True,
                )
                for i in range(params.down_sampling_layers + 1)
            ]
        )

        if (
            self.task_name == "long_term_forecast"
            or self.task_name == "short_term_forecast"
        ):
            self.predict_layers = torch.nn.ModuleList(
                [
                    torch.nn.Linear(
                        self.seq_len // (params.down_sampling_window**i),
                        self.pred_len,
                    )
                    for i in range(params.down_sampling_layers + 1)
                ]
            )

            if self.channel_independence == 1:
                self.projection_layer = nn.Linear(params.d_model, 1, bias=True)
            else:
                self.projection_layer = nn.Linear(
                    params.d_model, params.c_out, bias=True
                )

                self.out_res_layers = torch.nn.ModuleList(
                    [
                        torch.nn.Linear(
                            self.seq_len // (params.down_sampling_window**i),
                            self.seq_len // (params.down_sampling_window**i),
                        )
                        for i in range(params.down_sampling_layers + 1)
                    ]
                )

                self.regression_layers = torch.nn.ModuleList(
                    [
                        torch.nn.Linear(
                            self.seq_len // (params.down_sampling_window**i),
                            self.pred_len,
                        )
                        for i in range(params.down_sampling_layers + 1)
                    ]
                )
        if self.task_name == "imputation" or self.task_name == "anomaly_detection":
            if self.channel_independence == 1:
                self.projection_layer = nn.Linear(params.d_model, 1, bias=True)
            else:
                self.projection_layer = nn.Linear(
                    params.d_model, params.c_out, bias=True
                )
        if self.task_name == "classification":
            self.act = F.gelu
            self.dropout = nn.Dropout(params.dropout)
            self.projection = nn.Linear(params.d_model * self.seq_len, params.num_class)

    def out_projection(self, dec_out, i, out_res):
        dec_out = self.projection_layer(dec_out)
        out_res = out_res.permute(0, 2, 1)
        out_res = self.out_res_layers[i](out_res)
        out_res = self.regression_layers[i](out_res).permute(0, 2, 1)
        dec_out = dec_out + out_res
        return dec_out

    def pre_enc(self, x_list):
        if self.channel_independence == 1:
            return (x_list, None)
        else:
            out1_list = []
            out2_list = []
            for x in x_list:
                x_1, x_2 = self.preprocess(x)
                out1_list.append(x_1)
                out2_list.append(x_2)
            return (out1_list, out2_list)

    def __multi_scale_process_inputs(self, x_enc, x_mark_enc):
        if self.params.down_sampling_method == "max":
            down_pool = torch.nn.MaxPool1d(
                self.params.down_sampling_window, return_indices=False
            )
        elif self.params.down_sampling_method == "avg":
            down_pool = torch.nn.AvgPool1d(self.params.down_sampling_window)
        elif self.params.down_sampling_method == "conv":
            padding = 1 if torch.__version__ >= "1.5.0" else 2
            down_pool = nn.Conv1d(
                in_channels=self.params.enc_in,
                out_channels=self.params.enc_in,
                kernel_size=3,
                padding=padding,
                stride=self.params.down_sampling_window,
                padding_mode="circular",
                bias=False,
            )
        else:
            return x_enc, x_mark_enc
        # B,T,C -> B,C,T
        x_enc = x_enc.permute(0, 2, 1)

        x_enc_ori = x_enc
        x_mark_enc_mark_ori = x_mark_enc

        x_enc_sampling_list = []
        x_mark_sampling_list = []
        x_enc_sampling_list.append(x_enc.permute(0, 2, 1))
        x_mark_sampling_list.append(x_mark_enc)

        for i in range(self.params.down_sampling_layers):
            x_enc_sampling = down_pool(x_enc_ori)

            x_enc_sampling_list.append(x_enc_sampling.permute(0, 2, 1))
            x_enc_ori = x_enc_sampling

            if x_mark_enc_mark_ori is not None:
                x_mark_sampling_list.append(
                    x_mark_enc_mark_ori[:, :: self.params.down_sampling_window, :]
                )
                x_mark_enc_mark_ori = x_mark_enc_mark_ori[
                    :, :: self.params.down_sampling_window, :
                ]

        x_enc = x_enc_sampling_list
        if x_mark_enc_mark_ori is not None:
            x_mark_enc = x_mark_sampling_list
        else:
            x_mark_enc = x_mark_enc

        return x_enc, x_mark_enc

    def forecast(self, x_enc, x_mark_enc):

        # if self.use_future_temporal_feature:
        #     if self.channel_independence == 1:
        #         B, T, N = x_enc.size()
        #         x_mark_dec = x_mark_dec.repeat(N, 1, 1)
        #         self.x_mark_dec = self.enc_embedding(None, x_mark_dec)
        #     else:
        #         self.x_mark_dec = self.enc_embedding(None, x_mark_dec)

        x_enc, x_mark_enc = self.__multi_scale_process_inputs(x_enc, x_mark_enc)

        x_list = []
        x_mark_list = []
        if x_mark_enc is not None:
            for i, x, x_mark in zip(range(len(x_enc)), x_enc, x_mark_enc):
                B, T, N = x.size()
                x = self.normalize_layers[i](x, "norm")
                if self.channel_independence == 1:
                    x = x.permute(0, 2, 1).contiguous().reshape(B * N, T, 1)
                    x_mark = x_mark.repeat(N, 1, 1)
                x_list.append(x)
                x_mark_list.append(x_mark)
        else:
            for i, x in zip(
                range(len(x_enc)),
                x_enc,
            ):
                B, T, N = x.size()
                x = self.normalize_layers[i](x, "norm")
                if self.channel_independence == 1:
                    x = x.permute(0, 2, 1).contiguous().reshape(B * N, T, 1)
                x_list.append(x)

        # embedding
        enc_out_list = []
        x_list = self.pre_enc(x_list)
        if x_mark_enc is not None:
            for i, x, x_mark in zip(range(len(x_list[0])), x_list[0], x_mark_list):
                enc_out = self.enc_embedding(x, x_mark)  # [B,T,C]
                enc_out_list.append(enc_out)
        else:
            for i, x in zip(range(len(x_list[0])), x_list[0]):
                enc_out = self.enc_embedding(x, None)  # [B,T,C]
                enc_out_list.append(enc_out)

        # Past Decomposable Mixing as encoder for past
        for i in range(self.layer):
            enc_out_list = self.pdm_blocks[i](enc_out_list)

        # Future Multipredictor Mixing as decoder for future
        dec_out_list = self.future_multi_mixing(B, enc_out_list, x_list)

        dec_out = torch.stack(dec_out_list, dim=-1).sum(-1)
        dec_out = self.normalize_layers[0](dec_out, "denorm")
        return dec_out

    def future_multi_mixing(self, B, enc_out_list, x_list):
        dec_out_list = []
        if self.channel_independence == 1:
            x_list = x_list[0]
            for i, enc_out in zip(range(len(x_list)), enc_out_list):
                dec_out = self.predict_layers[i](enc_out.permute(0, 2, 1)).permute(
                    0, 2, 1
                )  # align temporal dimension
                if self.use_future_temporal_feature:
                    dec_out = dec_out + self.x_mark_dec
                    dec_out = self.projection_layer(dec_out)
                else:
                    dec_out = self.projection_layer(dec_out)
                dec_out = (
                    dec_out.reshape(B, self.params.c_out, self.pred_len)
                    .permute(0, 2, 1)
                    .contiguous()
                )
                dec_out_list.append(dec_out)

        else:
            for i, enc_out, out_res in zip(
                range(len(x_list[0])), enc_out_list, x_list[1]
            ):
                dec_out = self.predict_layers[i](enc_out.permute(0, 2, 1)).permute(
                    0, 2, 1
                )  # align temporal dimension
                dec_out = self.out_projection(dec_out, i, out_res)
                dec_out_list.append(dec_out)

        return dec_out_list

    def classification(self, x_enc, x_mark_enc):
        x_enc, _ = self.__multi_scale_process_inputs(x_enc, None)
        x_list = x_enc

        # embedding
        enc_out_list = []
        for x in x_list:
            enc_out = self.enc_embedding(x, None)  # [B,T,C]
            enc_out_list.append(enc_out)

        # MultiScale-CrissCrossAttention  as encoder for past
        for i in range(self.layer):
            enc_out_list = self.pdm_blocks[i](enc_out_list)

        enc_out = enc_out_list[0]
        # Output
        # the output transformer encoder/decoder embeddings don't include non-linearity
        output = self.act(enc_out)
        output = self.dropout(output)
        # zero-out padding embeddings
        output = output * x_mark_enc.unsqueeze(-1)
        # (batch_size, seq_length * d_model)
        output = output.reshape(output.shape[0], -1)
        output = self.projection(output)  # (batch_size, num_classes)
        return output

    def anomaly_detection(self, x_enc):
        B, T, N = x_enc.size()
        x_enc, _ = self.__multi_scale_process_inputs(x_enc, None)

        x_list = []

        for i, x in zip(
            range(len(x_enc)),
            x_enc,
        ):
            B, T, N = x.size()
            x = self.normalize_layers[i](x, "norm")
            if self.channel_independence == 1:
                x = x.permute(0, 2, 1).contiguous().reshape(B * N, T, 1)
            x_list.append(x)

        # embedding
        enc_out_list = []
        for x in x_list:
            enc_out = self.enc_embedding(x, None)  # [B,T,C]
            enc_out_list.append(enc_out)

        # MultiScale-CrissCrossAttention  as encoder for past
        for i in range(self.layer):
            enc_out_list = self.pdm_blocks[i](enc_out_list)

        dec_out = self.projection_layer(enc_out_list[0])
        dec_out = (
            dec_out.reshape(B, self.params.c_out, -1).permute(0, 2, 1).contiguous()
        )

        dec_out = self.normalize_layers[0](dec_out, "denorm")
        return dec_out

    def imputation(self, x_enc, x_mark_enc, mask):
        means = torch.sum(x_enc, dim=1) / torch.sum(mask == 1, dim=1)
        means = means.unsqueeze(1).detach()
        x_enc = x_enc - means
        x_enc = x_enc.masked_fill(mask == 0, 0)
        stdev = torch.sqrt(
            torch.sum(x_enc * x_enc, dim=1) / torch.sum(mask == 1, dim=1) + 1e-5
        )
        stdev = stdev.unsqueeze(1).detach()
        x_enc /= stdev

        B, T, N = x_enc.size()
        x_enc, x_mark_enc = self.__multi_scale_process_inputs(x_enc, x_mark_enc)

        x_list = []
        x_mark_list = []
        if x_mark_enc is not None:
            for i, x, x_mark in zip(range(len(x_enc)), x_enc, x_mark_enc):
                B, T, N = x.size()
                if self.channel_independence == 1:
                    x = x.permute(0, 2, 1).contiguous().reshape(B * N, T, 1)
                x_list.append(x)
                x_mark = x_mark.repeat(N, 1, 1)
                x_mark_list.append(x_mark)
        else:
            for i, x in zip(
                range(len(x_enc)),
                x_enc,
            ):
                B, T, N = x.size()
                if self.channel_independence == 1:
                    x = x.permute(0, 2, 1).contiguous().reshape(B * N, T, 1)
                x_list.append(x)

        # embedding
        enc_out_list = []
        for x in x_list:
            enc_out = self.enc_embedding(x, None)  # [B,T,C]
            enc_out_list.append(enc_out)

        # MultiScale-CrissCrossAttention  as encoder for past
        for i in range(self.layer):
            enc_out_list = self.pdm_blocks[i](enc_out_list)

        dec_out = self.projection_layer(enc_out_list[0])
        dec_out = (
            dec_out.reshape(B, self.params.c_out, -1).permute(0, 2, 1).contiguous()
        )

        dec_out = dec_out * (stdev[:, 0, :].unsqueeze(1).repeat(1, self.seq_len, 1))
        dec_out = dec_out + (means[:, 0, :].unsqueeze(1).repeat(1, self.seq_len, 1))
        return dec_out

    def forward(self, x_enc, x_mark_enc, mask=None):
        x_mark_enc = None
        if (
            self.task_name == "long_term_forecast"
            or self.task_name == "short_term_forecast"
        ):
            dec_out = self.forecast(x_enc, x_mark_enc)
            return dec_out
        if self.task_name == "imputation":
            dec_out = self.imputation(x_enc, x_mark_enc, mask)
            return dec_out  # [B, L, D]
        if self.task_name == "anomaly_detection":
            dec_out = self.anomaly_detection(x_enc)
            return dec_out  # [B, L, D]
        if self.task_name == "classification":
            dec_out = self.classification(x_enc, x_mark_enc)
            return dec_out  # [B, N]
        else:
            raise ValueError("Other tasks implemented yet")

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        pad_mask = batch.data[:, : self.seq_len]
        y_hat = self(x, pad_mask)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x, pad_mask)
        return y_hat


if __name__ == "__main__":
    m = TimeMixer(10, 5, 2, output_channels=4)
    x = torch.randn(3, 10, 2)
    y = m(x)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor, nn

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.0
    batch_size: int = 32
    lr: float = 0.005
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path: str = f"logs"
    activation_fn: str = "relu"
    blocks: int = 4
    num_blocks: int = 4
    ff_dim: int = 256
    output_channels: int = None
    norm_type: str = "batch"
    normalize_before: bool = True
    revin: bool = True
    affine: bool = False
    subtract_last: bool = False


def get_mutlivariate_params():
    params = Hyperparams(
        optimizer=torch.optim.Adam,
        loss_fn=torch.nn.MSELoss(),
        batch_size=32,
        batches_per_epoch=250,
        training_method=TrainingMethod.DMS,
        model_file_path="logs",
        epochs=10,
        scheduler=Scheduler.STEP_LR,
        lr=0.001,
        weight_decay=0.0,
        blocks=4,
        dropout=0.9,
        num_blocks=4,
        ff_dim=256,
        output_channels=7,
    )
    return params


class RevIN(nn.Module):
    def __init__(self, num_features: int, eps=1e-5, affine=True, subtract_last=False):
        """
        :param num_features: the number of features or channels
        :param eps: a value added for numerical stability
        :param affine: if True, RevIN has learnable affine parameters
        """
        super(RevIN, self).__init__()
        self.num_features = num_features
        self.eps = eps
        self.affine = affine
        self.subtract_last = subtract_last
        if self.affine:
            self._init_params()

    def forward(self, x, mode: str):
        if mode == "norm":
            self._get_statistics(x)
            x = self._normalize(x)
        elif mode == "denorm":
            x = self._denormalize(x)
        else:
            raise NotImplementedError
        return x

    def _init_params(self):
        # initialize RevIN params: (C,)
        self.affine_weight = nn.Parameter(torch.ones(self.num_features))
        self.affine_bias = nn.Parameter(torch.zeros(self.num_features))

    def _get_statistics(self, x):
        dim2reduce = tuple(range(1, x.ndim - 1))
        if self.subtract_last:
            self.last = x[:, -1, :].unsqueeze(1)
        else:
            self.mean = torch.mean(x, dim=dim2reduce, keepdim=True).detach()
        self.stdev = torch.sqrt(
            torch.var(x, dim=dim2reduce, keepdim=True, unbiased=False) + self.eps
        ).detach()

    def _normalize(self, x):
        if self.subtract_last:
            x = x - self.last
        else:
            x = x - self.mean
        x = x / self.stdev
        if self.affine:
            x = x * self.affine_weight
            x = x + self.affine_bias
        return x

    def _denormalize(self, x):
        if self.affine:
            x = x - self.affine_bias
            x = x / (self.affine_weight + self.eps * self.eps)
        x = x * self.stdev
        if self.subtract_last:
            x = x + self.last
        else:
            x = x + self.mean
        return x


class TimeBatchNorm2d(nn.BatchNorm1d):
    """A batch normalization layer that normalizes over the last two dimensions of a
    sequence in PyTorch, mimicking Keras behavior.

    This class extends nn.BatchNorm1d to apply batch normalization across time and
    feature dimensions.

    Attributes:
        num_time_steps (int): Number of time steps in the input.
        num_channels (int): Number of channels in the input.
    """

    def __init__(self, normalized_shape: tuple[int, int]):
        """Initializes the TimeBatchNorm2d module.

        Args:
            normalized_shape (tuple[int, int]): A tuple (num_time_steps, num_channels)
                representing the shape of the time and feature dimensions to normalize.
        """
        num_time_steps, num_channels = normalized_shape
        super().__init__(num_channels * num_time_steps)
        self.num_time_steps = num_time_steps
        self.num_channels = num_channels

    def forward(self, x: Tensor) -> Tensor:
        """Applies the batch normalization over the last two dimensions of the input tensor.

        Args:
            x (Tensor): A 3D tensor with shape (N, S, C), where N is the batch size,
                S is the number of time steps, and C is the number of channels.

        Returns:
            Tensor: A 3D tensor with batch normalization applied over the last two dims.

        Raises:
            ValueError: If the input tensor is not 3D.
        """
        if x.ndim != 3:
            raise ValueError(
                f"Expected 3D input tensor, but got {x.ndim}D tensor instead."
            )

        # Reshaping input to combine time and feature dimensions for normalization
        x = x.reshape(x.shape[0], -1, 1)

        # Applying batch normalization
        x = super().forward(x)

        # Reshaping back to original dimensions (N, S, C)
        x = x.reshape(x.shape[0], self.num_time_steps, self.num_channels)

        return x


class FeatureMixing(nn.Module):
    """A module for feature mixing with flexibility in normalization and activation.

    This module provides options for batch normalization before or after mixing features,
    uses dropout for regularization, and allows for different activation functions.

    Args:
        sequence_length: The length of the sequences to be transformed.
        input_channels: The number of input channels to the module.
        output_channels: The number of output channels from the module.
        ff_dim: The dimension of the feed-forward network internal to the module.
        activation_fn: The activation function used within the feed-forward network.
        dropout_rate: The dropout probability used for regularization.
        normalize_before: A boolean indicating whether to apply normalization before
            the rest of the operations.
    """

    def __init__(
        self,
        sequence_length: int,
        input_channels: int,
        output_channels: int,
        ff_dim: int,
        activation_fn: Callable[[torch.Tensor], torch.Tensor] = F.relu,
        dropout_rate: float = 0.1,
        normalize_before: bool = True,
        norm_type: type[nn.Module] = TimeBatchNorm2d,
    ):
        """Initializes the FeatureMixing module with the provided parameters."""
        super().__init__()

        self.norm_before = (
            norm_type((sequence_length, input_channels))
            if normalize_before
            else nn.Identity()
        )
        self.norm_after = (
            norm_type((sequence_length, output_channels))
            if not normalize_before
            else nn.Identity()
        )

        self.activation_fn = activation_fn
        self.dropout = nn.Dropout(dropout_rate)
        self.fc1 = nn.Linear(input_channels, ff_dim)
        self.fc2 = nn.Linear(ff_dim, output_channels)

        self.projection = (
            nn.Linear(input_channels, output_channels)
            if input_channels != output_channels
            else nn.Identity()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the FeatureMixing module.

        Args:
            x: A 3D tensor with shape (N, C, L) where C is the channel dimension.

        Returns:
            The output tensor after feature mixing.
        """
        x_proj = self.projection(x)

        x = self.norm_before(x)

        x = self.fc1(x)  # Apply the first linear transformation.
        x = self.activation_fn(x)  # Apply the activation function.
        x = self.dropout(x)  # Apply dropout for regularization.
        x = self.fc2(x)  # Apply the second linear transformation.
        x = self.dropout(x)  # Apply dropout again if needed.

        x = x_proj + x  # Add the projection shortcut to the transformed features.

        return self.norm_after(x)


class ConditionalFeatureMixing(nn.Module):
    """Conditional feature mixing module that incorporates static features.

    This module extends the feature mixing process by including static features. It uses
    a linear transformation to integrate static features into the dynamic feature space,
    then applies the feature mixing on the concatenated features.

    Args:
        input_channels: The number of input channels of the dynamic features.
        output_channels: The number of output channels after feature mixing.
        static_channels: The number of channels in the static feature input.
        ff_dim: The inner dimension of the feedforward network used in feature mixing.
        activation_fn: The activation function used in feature mixing.
        dropout_rate: The dropout probability used in the feature mixing operation.
    """

    def __init__(
        self,
        sequence_length: int,
        input_channels: int,
        output_channels: int,
        static_channels: int,
        ff_dim: int,
        activation_fn: Callable = F.relu,
        dropout_rate: float = 0.1,
        normalize_before: bool = False,
        norm_type: type[nn.Module] = nn.LayerNorm,
    ):
        super().__init__()

        self.fr_static = nn.Linear(static_channels, output_channels)
        self.fm = FeatureMixing(
            sequence_length,
            input_channels + output_channels,
            output_channels,
            ff_dim,
            activation_fn,
            dropout_rate,
            normalize_before=normalize_before,
            norm_type=norm_type,
        )

    def forward(
        self, x: torch.Tensor, x_static: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Applies conditional feature mixing using both dynamic and static inputs.

        Args:
            x: A tensor representing dynamic features, typically with shape
               [batch_size, time_steps, input_channels].
            x_static: A tensor representing static features, typically with shape
               [batch_size, static_channels].

        Returns:
            A tuple containing:
            - The output tensor after applying conditional feature mixing.
            - The transformed static features tensor for monitoring or further processing.
        """
        v = self.fr_static(
            x_static
        )  # Transform static features to match output channels.
        v = v.unsqueeze(1).repeat(
            1, x.shape[1], 1
        )  # Repeat static features across time steps.

        return (
            self.fm(
                torch.cat([x, v], dim=-1)
            ),  # Apply feature mixing on concatenated features.
            v.detach(),  # Return detached static feature for monitoring or further use.
        )


class TimeMixing(nn.Module):
    """Applies a transformation over the time dimension of a sequence.

    This module applies a linear transformation followed by an activation function
    and dropout over the sequence length of the input feature tensor after converting
    feature maps to the time dimension and then back.

    Args:
        input_channels: The number of input channels to the module.
        sequence_length: The length of the sequences to be transformed.
        activation_fn: The activation function to be used after the linear transformation.
        dropout_rate: The dropout probability to be used after the activation function.
    """

    def __init__(
        self,
        sequence_length: int,
        input_channels: int,
        activation_fn: Callable = F.relu,
        dropout_rate: float = 0.1,
        norm_type: type[nn.Module] = TimeBatchNorm2d,
    ):
        """Initializes the TimeMixing module with the specified parameters."""
        super().__init__()
        self.norm = norm_type((sequence_length, input_channels))
        self.activation_fn = activation_fn
        self.dropout = nn.Dropout(dropout_rate)
        self.fc1 = nn.Linear(sequence_length, sequence_length)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Applies the time mixing operations on the input tensor.

        Args:
            x: A 3D tensor with shape (N, C, L), where C = channel dimension and
                L = sequence length.

        Returns:
            The normalized output tensor after time mixing transformations.
        """
        x_temp = feature_to_time(
            x
        )  # Convert feature maps to time dimension. Assumes definition elsewhere.
        x_temp = self.activation_fn(self.fc1(x_temp))
        x_temp = self.dropout(x_temp)
        x_res = time_to_feature(x_temp)  # Convert back from time to feature maps.

        return self.norm(
            x + x_res
        )  # Apply normalization and combine with original input.


class MixerLayer(nn.Module):
    """A residual block that combines time and feature mixing for sequence data.

    This module sequentially applies time mixing and feature mixing, which are forms
    of data augmentation and feature transformation that can help in learning temporal
    dependencies and feature interactions respectively.

    Args:
        sequence_length: The length of the input sequences.
        input_channels: The number of input channels to the module.
        output_channels: The number of output channels from the module.
        ff_dim: The inner dimension of the feedforward network used in feature mixing.
        activation_fn: The activation function used in both time and feature mixing.
        dropout_rate: The dropout probability used in both mixing operations.
    """

    def __init__(
        self,
        sequence_length: int,
        input_channels: int,
        output_channels: int,
        ff_dim: int,
        activation_fn: Callable = F.relu,
        dropout_rate: float = 0.1,
        normalize_before: bool = False,
        norm_type: type[nn.Module] = nn.LayerNorm,
    ):
        """Initializes the MixLayer with time and feature mixing modules."""
        super().__init__()

        self.time_mixing = TimeMixing(
            sequence_length,
            input_channels,
            activation_fn,
            dropout_rate,
            norm_type=norm_type,
        )
        self.feature_mixing = FeatureMixing(
            sequence_length,
            input_channels,
            output_channels,
            ff_dim,
            activation_fn,
            dropout_rate,
            norm_type=norm_type,
            normalize_before=normalize_before,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the MixLayer module.

        Args:
            x: A 3D tensor with shape (N, C, L) to be processed by the mixing layers.

        Returns:
            The output tensor after applying time and feature mixing operations.
        """
        x = self.time_mixing(x)  # Apply time mixing first.
        x = self.feature_mixing(x)  # Then apply feature mixing.

        return x


class ConditionalMixerLayer(nn.Module):
    """Conditional mix layer combining time and feature mixing with static context.

    This module combines time mixing and conditional feature mixing, where the latter
    is influenced by static features. This allows the module to learn representations
    that are influenced by both dynamic and static features.

    Args:
        sequence_length: The length of the input sequences.
        input_channels: The number of input channels of the dynamic features.
        output_channels: The number of output channels after feature mixing.
        static_channels: The number of channels in the static feature input.
        ff_dim: The inner dimension of the feedforward network used in feature mixing.
        activation_fn: The activation function used in both mixing operations.
        dropout_rate: The dropout probability used in both mixing operations.
    """

    def __init__(
        self,
        sequence_length: int,
        input_channels: int,
        output_channels: int,
        static_channels: int,
        ff_dim: int,
        activation_fn: Callable = F.relu,
        dropout_rate: float = 0.1,
        normalize_before: bool = False,
        norm_type: type[nn.Module] = nn.LayerNorm,
    ):
        super().__init__()

        self.time_mixing = TimeMixing(
            sequence_length,
            input_channels,
            activation_fn,
            dropout_rate,
            norm_type=norm_type,
        )
        self.feature_mixing = ConditionalFeatureMixing(
            sequence_length,
            input_channels,
            output_channels=output_channels,
            static_channels=static_channels,
            ff_dim=ff_dim,
            activation_fn=activation_fn,
            dropout_rate=dropout_rate,
            normalize_before=normalize_before,
            norm_type=norm_type,
        )

    def forward(self, x: torch.Tensor, x_static: torch.Tensor) -> torch.Tensor:
        """Forward pass for the conditional mix layer.

        Args:
            x: A tensor representing dynamic features, typically with shape
               [batch_size, time_steps, input_channels].
            x_static: A tensor representing static features, typically with shape
               [batch_size, static_channels].

        Returns:
            The output tensor after applying time and conditional feature mixing.
        """
        x = self.time_mixing(x)  # Apply time mixing first.
        x, _ = self.feature_mixing(
            x, x_static
        )  # Then apply conditional feature mixing.

        return x


def time_to_feature(x: torch.Tensor) -> torch.Tensor:
    """Converts a time series tensor to a feature tensor."""
    return x.permute(0, 2, 1)


feature_to_time = time_to_feature


class TSMixer(nn.Module):
    """TSMixer model for time series forecasting.

    This model uses a series of mixer layers to process time series data,
    followed by a linear transformation to project the output to the desired
    prediction length.

    Attributes:
        mixer_layers: Sequential container of mixer layers.
        temporal_projection: Linear layer for temporal projection.

    Args:
        sequence_length: Length of the input time series sequence.
        prediction_length: Desired length of the output prediction sequence.
        input_channels: Number of input channels.
        output_channels: Number of output channels. Defaults to None.
        activation_fn: Activation function to use. Defaults to "relu".
        num_blocks: Number of mixer blocks. Defaults to 2.
        dropout_rate: Dropout rate for regularization. Defaults to 0.1.
        ff_dim: Dimension of feedforward network inside mixer layer. Defaults to 64.
        normalize_before: Whether to apply layer normalization before or after mixer layer.
        norm_type: Type of normalization to use. "batch" or "layer". Defaults to "batch".
    """

    def __init__(
        self,
        sequence_length: int,
        prediction_length: int,
        input_channels: int,
        params: Hyperparams,
    ):
        super().__init__()
        self.seq_len = sequence_length

        # Transform activation_fn to callable
        activation_fn = getattr(F, params.activation_fn)

        # Transform norm_type to callable
        assert params.norm_type in {
            "batch",
            "layer",
        }, f"Invalid norm_type: {params.norm_type}, must be one of batch, layer."
        norm_type = TimeBatchNorm2d if params.norm_type == "batch" else nn.LayerNorm

        # Build mixer layers
        self.mixer_layers = self._build_mixer(
            params.num_blocks,
            input_channels,
            params.output_channels,
            ff_dim=params.ff_dim,
            activation_fn=activation_fn,
            dropout_rate=params.dropout,
            sequence_length=sequence_length,
            normalize_before=params.normalize_before,
            norm_type=norm_type,
        )

        # Temporal projection layer
        self.temporal_projection = nn.Linear(sequence_length, prediction_length)
        self.revin = params.revin
        if self.revin:
            self.revin_layer = RevIN(
                input_channels, affine=params.affine, subtract_last=params.subtract_last
            )

    def _build_mixer(
        self, num_blocks: int, input_channels: int, output_channels: int, **kwargs
    ):
        """Build the mixer blocks for the model.

        Args:
            num_blocks (int): Number of mixer blocks to be built.
            input_channels (int): Number of input channels for the first block.
            output_channels (int): Number of output channels for the last block.
            **kwargs: Additional keyword arguments for mixer layer configuration.

        Returns:
            nn.Sequential: Sequential container of mixer layers.
        """
        output_channels = (
            output_channels if output_channels is not None else input_channels
        )
        channels = [input_channels] * (num_blocks - 1) + [output_channels]

        return nn.Sequential(
            *[
                MixerLayer(input_channels=in_ch, output_channels=out_ch, **kwargs)
                for in_ch, out_ch in zip(channels[:-1], channels[1:])
            ]
        )

    def forward(self, x_hist: torch.Tensor) -> torch.Tensor:
        """Forward pass of the TSMixer model.

        Args:
            x_hist (torch.Tensor): Input time series tensor.

        Returns:
            torch.Tensor: The output tensor after processing by the model.
        """
        # normalise
        if self.revin:
            x_hist = self.revin_layer(x_hist, "norm")

        x = self.mixer_layers(x_hist)

        x_temp = feature_to_time(x)
        x_temp = self.temporal_projection(x_temp)
        x = time_to_feature(x_temp)

        # denorm
        if self.revin:
            x = self.revin_layer(x, "denorm")

        return x

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        y_hat = self(x)
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x)
        return y_hat


if __name__ == "__main__":
    m = TSMixer(10, 5, 2, output_channels=4)
    x = torch.randn(3, 10, 2)
    y = m(x)



from abc import ABC, abstractmethod


class Listener(ABC):
    @abstractmethod
    def notify(self, event):
        raise NotImplementedError


class EventManager:
    def __init__(self):
        self.listeners = {}

    def add_listener(self, event_type, listener):
        if event_type not in self.listeners:
            self.listeners[event_type] = []
        self.listeners[event_type].append(listener)

    def remove_listener(self, event_type, listener):
        if event_type in self.listeners:
            self.listeners[event_type].remove(listener)

    def notify(self, event):
        event_type = event.__class__
        if event_type in self.listeners:
            for listener in self.listeners[event_type]:
                listener.notify(event)



from dataclasses import dataclass
from typing import Any, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:
    """Class for keeping track of training and model params."""

    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.l1_loss
    dropout: float = 0.0
    batch_size: int = 32
    lr: float = 0.005
    epochs: int = 100
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = 50
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.REDUCE_LR_ON_PLATEAU
    model_file_path: str = f"logs"
    individual: bool = True
    kernel_size = 25
    scaling_fn: Any = None
    enc_in: int = 1


def get_mutlivariate_params():
    params = Hyperparams(
        optimizer=torch.optim.Adam,
        loss_fn=torch.nn.MSELoss(),
        batch_size=32,
        batches_per_epoch=100,
        training_method=TrainingMethod.DMS,
        model_file_path="logs",
        epochs=10,
        scheduler=Scheduler.STEP_LR,
        lr=0.005,
        weight_decay=0.0,
    )
    return params


class NLinear(nn.Module):
    """
    Normalization-Linear
    """

    def __init__(self, h: int, input_size: int, c_in: int, configs: Hyperparams):
        super(NLinear, self).__init__()
        print("enc_in", configs.enc_in)
        self.seq_len = input_size
        self.pred_len = h
        self.scaling_fn = configs.scaling_fn

        # Use this line if you want to visualize the weights
        # self.Linear.weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
        self.channels = c_in
        self.individual = configs.individual
        if self.individual:
            self.Linear = nn.ModuleList()
            for i in range(self.channels):
                self.Linear.append(nn.Linear(self.seq_len, self.pred_len))
        else:
            self.Linear = nn.Linear(self.seq_len, self.pred_len)

    def forward(self, x):
        # x: [Batch, Input length, Channel]
        seq_last = x[:, -1:, :].detach()
        x = x - seq_last
        if self.individual:
            output = torch.zeros(
                [x.size(0), self.pred_len, x.size(2)], dtype=x.dtype
            ).to(x.device)
            for i in range(self.channels):
                output[:, :, i] = self.Linear[i](x[:, :, i])
            x = output
        else:
            x = self.Linear(x.permute(0, 2, 1)).permute(0, 2, 1)
        x = x + seq_last
        return x  # [Batch, Output length, Channel]

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data
        y_hat = self(x[:, : self.seq_len, :])
        y = x[:, self.seq_len :, :]
        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x)
        return y_hat



import math
from dataclasses import dataclass, field
from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from nnts.utils import Scheduler, TrainingMethod

from ..datasets import PaddedData


@dataclass
class Hyperparams:

    optimizer: callable = torch.optim.Adam
    loss_fn: callable = F.mse_loss
    dropout: float = field(default=0.05, metadata={"help": "dropout"})
    batch_size: int = 128
    lr: float = 0.0001  # for one cycle
    epochs: int = 50
    patience: int = 10
    early_stopper_patience: int = 30
    batches_per_epoch: int = None
    weight_decay: float = 0.0
    training_method: TrainingMethod = TrainingMethod.DMS
    scheduler: Scheduler = Scheduler.ONE_CYCLE
    model_file_path = f"logs"

    output_attention: bool = field(
        default=False, metadata={"help": "whether to output attention in encoder"}
    )
    embed_type: int = 3
    freq: str = field(default="h", metadata={"help": "freq for time features encoding"})

    # Model definition Benchmarked
    # d_model: int = field(default=16, metadata={"help": "dimension of model"})
    # n_heads: int = field(default=4, metadata={"help": "num of heads"})
    # e_layers: int = field(default=2, metadata={"help": "num of encoder layers"})
    # d_layers: int = field(default=1, metadata={"help": "num of decoder layers"})
    # d_ff: int = field(default=128, metadata={"help": "dimension of fcn"})

    # Model definition Papers
    d_model: int = field(default=512, metadata={"help": "dimension of model"})
    n_heads: int = field(default=8, metadata={"help": "num of heads"})
    e_layers: int = field(default=2, metadata={"help": "num of encoder layers"})
    d_layers: int = field(default=1, metadata={"help": "num of decoder layers"})
    d_ff: int = field(default=2048, metadata={"help": "dimension of fcn"})

    moving_avg: int = field(
        default=25, metadata={"help": "window size of moving average"}
    )
    factor: int = field(default=1, metadata={"help": "attn factor"})
    embed: str = field(default="timeF", metadata={"help": "time features encoding"})
    activation: str = field(default="gelu", metadata={"help": "activation"})
    output_attention: bool = field(
        default=False, metadata={"help": "whether to output attention in encoder"}
    )
    do_predict: bool = field(
        default=False, metadata={"help": "whether to predict unseen future data"}
    )

    # Optimization
    des: str = field(default="test", metadata={"help": "exp description"})
    revin: bool = True
    affine: bool = False
    subtract_last = False


class RevIN(nn.Module):
    def __init__(self, num_features: int, eps=1e-5, affine=True, subtract_last=False):
        """
        :param num_features: the number of features or channels
        :param eps: a value added for numerical stability
        :param affine: if True, RevIN has learnable affine parameters
        """
        super(RevIN, self).__init__()
        self.num_features = num_features
        self.eps = eps
        self.affine = affine
        self.subtract_last = subtract_last
        if self.affine:
            self._init_params()

    def forward(self, x, mode: str):
        if mode == "norm":
            self._get_statistics(x)
            x = self._normalize(x)
        elif mode == "denorm":
            x = self._denormalize(x)
        else:
            raise NotImplementedError
        return x

    def _init_params(self):
        # initialize RevIN params: (C,)
        self.affine_weight = nn.Parameter(torch.ones(self.num_features))
        self.affine_bias = nn.Parameter(torch.zeros(self.num_features))

    def _get_statistics(self, x):
        dim2reduce = tuple(range(1, x.ndim - 1))
        if self.subtract_last:
            self.last = x[:, -1, :].unsqueeze(1)
        else:
            self.mean = torch.mean(x, dim=dim2reduce, keepdim=True).detach()
        self.stdev = torch.sqrt(
            torch.var(x, dim=dim2reduce, keepdim=True, unbiased=False) + self.eps
        ).detach()

    def _normalize(self, x):
        if self.subtract_last:
            x = x - self.last
        else:
            x = x - self.mean
        x = x / self.stdev
        if self.affine:
            x = x * self.affine_weight
            x = x + self.affine_bias
        return x

    def _denormalize(self, x):
        if self.affine:
            x = x - self.affine_bias
            x = x / (self.affine_weight + self.eps * self.eps)
        x = x * self.stdev
        if self.subtract_last:
            x = x + self.last
        else:
            x = x + self.mean
        return x


class PositionalEmbedding(nn.Module):
    def __init__(self, d_model, max_len=5000):
        super(PositionalEmbedding, self).__init__()
        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model).float()
        pe.require_grad = False

        position = torch.arange(0, max_len).float().unsqueeze(1)
        div_term = (
            torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)
        ).exp()

        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)

        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        return self.pe[:, : x.size(1)]


class TokenEmbedding(nn.Module):
    def __init__(self, c_in, d_model):
        super(TokenEmbedding, self).__init__()
        padding = 1 if torch.__version__ >= "1.5.0" else 2
        self.tokenConv = nn.Conv1d(
            in_channels=c_in,
            out_channels=d_model,
            kernel_size=3,
            padding=padding,
            padding_mode="circular",
            bias=False,
        )
        for m in self.modules():
            if isinstance(m, nn.Conv1d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_in", nonlinearity="leaky_relu"
                )

    def forward(self, x):
        x = self.tokenConv(x.permute(0, 2, 1)).transpose(1, 2)
        return x


class FixedEmbedding(nn.Module):
    def __init__(self, c_in, d_model):
        super(FixedEmbedding, self).__init__()

        w = torch.zeros(c_in, d_model).float()
        w.require_grad = False

        position = torch.arange(0, c_in).float().unsqueeze(1)
        div_term = (
            torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)
        ).exp()

        w[:, 0::2] = torch.sin(position * div_term)
        w[:, 1::2] = torch.cos(position * div_term)

        self.emb = nn.Embedding(c_in, d_model)
        self.emb.weight = nn.Parameter(w, requires_grad=False)

    def forward(self, x):
        return self.emb(x).detach()


class TemporalEmbedding(nn.Module):
    def __init__(self, d_model, embed_type="fixed", freq="h"):
        super(TemporalEmbedding, self).__init__()

        minute_size = 4
        hour_size = 24
        weekday_size = 7
        day_size = 32
        month_size = 13

        Embed = FixedEmbedding if embed_type == "fixed" else nn.Embedding
        if freq == "t":
            self.minute_embed = Embed(minute_size, d_model)
        self.hour_embed = Embed(hour_size, d_model)
        self.weekday_embed = Embed(weekday_size, d_model)
        self.day_embed = Embed(day_size, d_model)
        self.month_embed = Embed(month_size, d_model)

    def forward(self, x):
        x = x.long()

        minute_x = (
            self.minute_embed(x[:, :, 4]) if hasattr(self, "minute_embed") else 0.0
        )
        hour_x = self.hour_embed(x[:, :, 3])
        weekday_x = self.weekday_embed(x[:, :, 2])
        day_x = self.day_embed(x[:, :, 1])
        month_x = self.month_embed(x[:, :, 0])

        return hour_x + weekday_x + day_x + month_x + minute_x


class TimeFeatureEmbedding(nn.Module):
    def __init__(self, d_model, embed_type="timeF", freq="h"):
        super(TimeFeatureEmbedding, self).__init__()

        freq_map = {"h": 4, "t": 5, "s": 6, "m": 1, "a": 1, "w": 2, "d": 3, "b": 3}
        d_inp = freq_map[freq]
        self.embed = nn.Linear(d_inp, d_model, bias=False)

    def forward(self, x):
        return self.embed(x)


class DataEmbedding(nn.Module):
    def __init__(self, c_in, d_model, embed_type="fixed", freq="h", dropout=0.1):
        super(DataEmbedding, self).__init__()

        self.value_embedding = TokenEmbedding(c_in=c_in, d_model=d_model)
        self.position_embedding = PositionalEmbedding(d_model=d_model)
        self.temporal_embedding = (
            TemporalEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
            if embed_type != "timeF"
            else TimeFeatureEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x, x_mark):
        x = (
            self.value_embedding(x)
            + self.temporal_embedding(x_mark)
            + self.position_embedding(x)
        )
        return self.dropout(x)


class DataEmbedding_wo_pos(nn.Module):
    def __init__(self, c_in, d_model, embed_type="fixed", freq="h", dropout=0.1):
        super(DataEmbedding_wo_pos, self).__init__()

        self.value_embedding = TokenEmbedding(c_in=c_in, d_model=d_model)
        self.position_embedding = PositionalEmbedding(d_model=d_model)
        self.temporal_embedding = (
            TemporalEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
            if embed_type != "timeF"
            else TimeFeatureEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x, x_mark):
        x = self.value_embedding(x) + self.temporal_embedding(x_mark)
        return self.dropout(x)


class DataEmbedding_wo_pos_temp(nn.Module):
    def __init__(self, c_in, d_model, embed_type="fixed", freq="h", dropout=0.1):
        super(DataEmbedding_wo_pos_temp, self).__init__()

        self.value_embedding = TokenEmbedding(c_in=c_in, d_model=d_model)
        self.position_embedding = PositionalEmbedding(d_model=d_model)
        self.temporal_embedding = (
            TemporalEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
            if embed_type != "timeF"
            else TimeFeatureEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x, x_mark):
        x = self.value_embedding(x)
        return self.dropout(x)


class DataEmbedding_wo_temp(nn.Module):
    def __init__(self, c_in, d_model, embed_type="fixed", freq="h", dropout=0.1):
        super(DataEmbedding_wo_temp, self).__init__()

        self.value_embedding = TokenEmbedding(c_in=c_in, d_model=d_model)
        self.position_embedding = PositionalEmbedding(d_model=d_model)
        self.temporal_embedding = (
            TemporalEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
            if embed_type != "timeF"
            else TimeFeatureEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
        )
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, x, x_mark):
        x = self.value_embedding(x) + self.position_embedding(x)
        return self.dropout(x)


class my_Layernorm(nn.Module):
    """
    Special designed layernorm for the seasonal part
    """

    def __init__(self, channels):
        super(my_Layernorm, self).__init__()
        self.layernorm = nn.LayerNorm(channels)

    def forward(self, x):
        x_hat = self.layernorm(x)
        bias = torch.mean(x_hat, dim=1).unsqueeze(1).repeat(1, x.shape[1], 1)
        return x_hat - bias


class moving_avg(nn.Module):
    """
    Moving average block to highlight the trend of time series
    """

    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class series_decomp(nn.Module):
    """
    Series decomposition block
    """

    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean


class EncoderLayer(nn.Module):
    """
    Autoformer encoder layer with the progressive decomposition architecture
    """

    def __init__(
        self,
        attention,
        d_model,
        d_ff=None,
        moving_avg=25,
        dropout=0.1,
        activation="relu",
    ):
        super(EncoderLayer, self).__init__()
        d_ff = d_ff or 4 * d_model
        self.attention = attention
        self.conv1 = nn.Conv1d(
            in_channels=d_model, out_channels=d_ff, kernel_size=1, bias=False
        )
        self.conv2 = nn.Conv1d(
            in_channels=d_ff, out_channels=d_model, kernel_size=1, bias=False
        )
        self.decomp1 = series_decomp(moving_avg)
        self.decomp2 = series_decomp(moving_avg)
        self.dropout = nn.Dropout(dropout)
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, attn_mask=None):
        new_x, attn = self.attention(x, x, x, attn_mask=attn_mask)
        x = x + self.dropout(new_x)
        x, _ = self.decomp1(x)
        y = x
        y = self.dropout(self.activation(self.conv1(y.transpose(-1, 1))))
        y = self.dropout(self.conv2(y).transpose(-1, 1))
        res, _ = self.decomp2(x + y)
        return res, attn


class Encoder(nn.Module):
    """
    Autoformer encoder
    """

    def __init__(self, attn_layers, conv_layers=None, norm_layer=None):
        super(Encoder, self).__init__()
        self.attn_layers = nn.ModuleList(attn_layers)
        self.conv_layers = (
            nn.ModuleList(conv_layers) if conv_layers is not None else None
        )
        self.norm = norm_layer

    def forward(self, x, attn_mask=None):
        attns = []
        if self.conv_layers is not None:
            for attn_layer, conv_layer in zip(self.attn_layers, self.conv_layers):
                x, attn = attn_layer(x, attn_mask=attn_mask)
                x = conv_layer(x)
                attns.append(attn)
            x, attn = self.attn_layers[-1](x)
            attns.append(attn)
        else:
            for attn_layer in self.attn_layers:
                x, attn = attn_layer(x, attn_mask=attn_mask)
                attns.append(attn)

        if self.norm is not None:
            x = self.norm(x)

        return x, attns


class DecoderLayer(nn.Module):
    """
    Autoformer decoder layer with the progressive decomposition architecture
    """

    def __init__(
        self,
        self_attention,
        cross_attention,
        d_model,
        c_out,
        d_ff=None,
        moving_avg=25,
        dropout=0.1,
        activation="relu",
    ):
        super(DecoderLayer, self).__init__()
        d_ff = d_ff or 4 * d_model
        self.self_attention = self_attention
        self.cross_attention = cross_attention
        self.conv1 = nn.Conv1d(
            in_channels=d_model, out_channels=d_ff, kernel_size=1, bias=False
        )
        self.conv2 = nn.Conv1d(
            in_channels=d_ff, out_channels=d_model, kernel_size=1, bias=False
        )
        self.decomp1 = series_decomp(moving_avg)
        self.decomp2 = series_decomp(moving_avg)
        self.decomp3 = series_decomp(moving_avg)
        self.dropout = nn.Dropout(dropout)
        self.projection = nn.Conv1d(
            in_channels=d_model,
            out_channels=c_out,
            kernel_size=3,
            stride=1,
            padding=1,
            padding_mode="circular",
            bias=False,
        )
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, cross, x_mask=None, cross_mask=None):
        x = x + self.dropout(self.self_attention(x, x, x, attn_mask=x_mask)[0])
        x, trend1 = self.decomp1(x)
        x = x + self.dropout(
            self.cross_attention(x, cross, cross, attn_mask=cross_mask)[0]
        )
        x, trend2 = self.decomp2(x)
        y = x
        y = self.dropout(self.activation(self.conv1(y.transpose(-1, 1))))
        y = self.dropout(self.conv2(y).transpose(-1, 1))
        x, trend3 = self.decomp3(x + y)

        residual_trend = trend1 + trend2 + trend3
        residual_trend = self.projection(residual_trend.permute(0, 2, 1)).transpose(
            1, 2
        )
        return x, residual_trend


class Decoder(nn.Module):
    """
    Autoformer encoder
    """

    def __init__(self, layers, norm_layer=None, projection=None):
        super(Decoder, self).__init__()
        self.layers = nn.ModuleList(layers)
        self.norm = norm_layer
        self.projection = projection

    def forward(self, x, cross, x_mask=None, cross_mask=None, trend=None):
        for layer in self.layers:
            x, residual_trend = layer(x, cross, x_mask=x_mask, cross_mask=cross_mask)
            trend = trend + residual_trend

        if self.norm is not None:
            x = self.norm(x)

        if self.projection is not None:
            x = self.projection(x)
        return x, trend


class AutoCorrelation(nn.Module):
    """
    AutoCorrelation Mechanism with the following two phases:
    (1) period-based dependencies discovery
    (2) time delay aggregation
    This block can replace the self-attention family mechanism seamlessly.
    """

    def __init__(
        self,
        mask_flag=True,
        factor=1,
        scale=None,
        attention_dropout=0.1,
        output_attention=False,
    ):
        super(AutoCorrelation, self).__init__()
        self.factor = factor
        self.scale = scale
        self.mask_flag = mask_flag
        self.output_attention = output_attention
        self.dropout = nn.Dropout(attention_dropout)

    def time_delay_agg_training(self, values, corr):
        """
        SpeedUp version of Autocorrelation (a batch-normalization style design)
        This is for the training phase.
        """
        head = values.shape[1]
        channel = values.shape[2]
        length = values.shape[3]
        # find top k
        top_k = int(self.factor * math.log(length))
        mean_value = torch.mean(torch.mean(corr, dim=1), dim=1)
        index = torch.topk(torch.mean(mean_value, dim=0), top_k, dim=-1)[1]
        weights = torch.stack([mean_value[:, index[i]] for i in range(top_k)], dim=-1)
        # update corr
        tmp_corr = torch.softmax(weights, dim=-1)
        # aggregation
        tmp_values = values
        delays_agg = torch.zeros_like(values).float()
        for i in range(top_k):
            pattern = torch.roll(tmp_values, -int(index[i]), -1)
            delays_agg = delays_agg + pattern * (
                tmp_corr[:, i]
                .unsqueeze(1)
                .unsqueeze(1)
                .unsqueeze(1)
                .repeat(1, head, channel, length)
            )
        return delays_agg

    def time_delay_agg_inference(self, values, corr):
        """
        SpeedUp version of Autocorrelation (a batch-normalization style design)
        This is for the inference phase.
        """
        batch = values.shape[0]
        head = values.shape[1]
        channel = values.shape[2]
        length = values.shape[3]
        # index init
        init_index = (
            torch.arange(length)
            .unsqueeze(0)
            .unsqueeze(0)
            .unsqueeze(0)
            .repeat(batch, head, channel, 1)
        )
        # find top k
        top_k = int(self.factor * math.log(length))
        mean_value = torch.mean(torch.mean(corr, dim=1), dim=1)
        weights = torch.topk(mean_value, top_k, dim=-1)[0]
        delay = torch.topk(mean_value, top_k, dim=-1)[1]
        # update corr
        tmp_corr = torch.softmax(weights, dim=-1)
        # aggregation
        tmp_values = values.repeat(1, 1, 1, 2)
        delays_agg = torch.zeros_like(values).float()
        for i in range(top_k):
            tmp_delay = init_index + delay[:, i].unsqueeze(1).unsqueeze(1).unsqueeze(
                1
            ).repeat(1, head, channel, length)
            pattern = torch.gather(tmp_values, dim=-1, index=tmp_delay)
            delays_agg = delays_agg + pattern * (
                tmp_corr[:, i]
                .unsqueeze(1)
                .unsqueeze(1)
                .unsqueeze(1)
                .repeat(1, head, channel, length)
            )
        return delays_agg

    def time_delay_agg_full(self, values, corr):
        """
        Standard version of Autocorrelation
        """
        batch = values.shape[0]
        head = values.shape[1]
        channel = values.shape[2]
        length = values.shape[3]
        # index init
        init_index = (
            torch.arange(length)
            .unsqueeze(0)
            .unsqueeze(0)
            .unsqueeze(0)
            .repeat(batch, head, channel, 1)
        )
        # find top k
        top_k = int(self.factor * math.log(length))
        weights = torch.topk(corr, top_k, dim=-1)[0]
        delay = torch.topk(corr, top_k, dim=-1)[1]
        # update corr
        tmp_corr = torch.softmax(weights, dim=-1)
        # aggregation
        tmp_values = values.repeat(1, 1, 1, 2)
        delays_agg = torch.zeros_like(values).float()
        for i in range(top_k):
            tmp_delay = init_index + delay[..., i].unsqueeze(-1)
            pattern = torch.gather(tmp_values, dim=-1, index=tmp_delay)
            delays_agg = delays_agg + pattern * (tmp_corr[..., i].unsqueeze(-1))
        return delays_agg

    def forward(self, queries, keys, values, attn_mask):
        B, L, H, E = queries.shape
        _, S, _, D = values.shape
        if L > S:
            zeros = torch.zeros_like(queries[:, : (L - S), :]).float()
            values = torch.cat([values, zeros], dim=1)
            keys = torch.cat([keys, zeros], dim=1)
        else:
            values = values[:, :L, :, :]
            keys = keys[:, :L, :, :]

        # period-based dependencies
        q_fft = torch.fft.rfft(queries.permute(0, 2, 3, 1).contiguous(), dim=-1)
        k_fft = torch.fft.rfft(keys.permute(0, 2, 3, 1).contiguous(), dim=-1)
        res = q_fft * torch.conj(k_fft)
        corr = torch.fft.irfft(res, dim=-1)

        # time delay agg
        if self.training:
            V = self.time_delay_agg_training(
                values.permute(0, 2, 3, 1).contiguous(), corr
            ).permute(0, 3, 1, 2)
        else:
            V = self.time_delay_agg_inference(
                values.permute(0, 2, 3, 1).contiguous(), corr
            ).permute(0, 3, 1, 2)

        if self.output_attention:
            return (V.contiguous(), corr.permute(0, 3, 1, 2))
        else:
            return (V.contiguous(), None)


class AutoCorrelationLayer(nn.Module):
    def __init__(self, correlation, d_model, n_heads, d_keys=None, d_values=None):
        super(AutoCorrelationLayer, self).__init__()

        d_keys = d_keys or (d_model // n_heads)
        d_values = d_values or (d_model // n_heads)

        self.inner_correlation = correlation
        self.query_projection = nn.Linear(d_model, d_keys * n_heads)
        self.key_projection = nn.Linear(d_model, d_keys * n_heads)
        self.value_projection = nn.Linear(d_model, d_values * n_heads)
        self.out_projection = nn.Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads

    def forward(self, queries, keys, values, attn_mask):
        B, L, _ = queries.shape
        _, S, _ = keys.shape
        H = self.n_heads

        queries = self.query_projection(queries).view(B, L, H, -1)
        keys = self.key_projection(keys).view(B, S, H, -1)
        values = self.value_projection(values).view(B, S, H, -1)

        out, attn = self.inner_correlation(queries, keys, values, attn_mask)
        out = out.view(B, L, -1)

        return self.out_projection(out), attn


class Autoformer(nn.Module):
    """
    Autoformer is the first method to achieve the series-wise connection,
    with inherent O(LlogL) complexity
    """

    def __init__(self, h: int, input_size: int, c_in: int, configs: Hyperparams):
        super(Autoformer, self).__init__()

        self.seq_len = input_size
        self.label_len = input_size
        self.pred_len = h
        self.output_attention = configs.output_attention
        self.c_in = c_in

        # Decomp
        kernel_size = configs.moving_avg
        self.decomp = series_decomp(kernel_size)

        # Embedding
        # The series-wise connection inherently contains the sequential information.
        # Thus, we can discard the position embedding of transformers.
        if configs.embed_type == 0:
            self.enc_embedding = DataEmbedding_wo_pos(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
            self.dec_embedding = DataEmbedding_wo_pos(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
        elif configs.embed_type == 1:
            self.enc_embedding = DataEmbedding(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
            self.dec_embedding = DataEmbedding(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
        elif configs.embed_type == 2:
            self.enc_embedding = DataEmbedding_wo_pos(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
            self.dec_embedding = DataEmbedding_wo_pos(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )

        elif configs.embed_type == 3:
            self.enc_embedding = DataEmbedding_wo_temp(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
            self.dec_embedding = DataEmbedding_wo_temp(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
        elif configs.embed_type == 4:
            self.enc_embedding = DataEmbedding_wo_pos_temp(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )
            self.dec_embedding = DataEmbedding_wo_pos_temp(
                c_in,
                configs.d_model,
                configs.embed,
                configs.freq,
                configs.dropout,
            )

        # Encoder
        self.encoder = Encoder(
            [
                EncoderLayer(
                    AutoCorrelationLayer(
                        AutoCorrelation(
                            False,
                            configs.factor,
                            attention_dropout=configs.dropout,
                            output_attention=configs.output_attention,
                        ),
                        configs.d_model,
                        configs.n_heads,
                    ),
                    configs.d_model,
                    configs.d_ff,
                    moving_avg=configs.moving_avg,
                    dropout=configs.dropout,
                    activation=configs.activation,
                )
                for l in range(configs.e_layers)
            ],
            norm_layer=my_Layernorm(configs.d_model),
        )
        # Decoder
        self.decoder = Decoder(
            [
                DecoderLayer(
                    AutoCorrelationLayer(
                        AutoCorrelation(
                            True,
                            configs.factor,
                            attention_dropout=configs.dropout,
                            output_attention=False,
                        ),
                        configs.d_model,
                        configs.n_heads,
                    ),
                    AutoCorrelationLayer(
                        AutoCorrelation(
                            False,
                            configs.factor,
                            attention_dropout=configs.dropout,
                            output_attention=False,
                        ),
                        configs.d_model,
                        configs.n_heads,
                    ),
                    configs.d_model,
                    c_in,
                    configs.d_ff,
                    moving_avg=configs.moving_avg,
                    dropout=configs.dropout,
                    activation=configs.activation,
                )
                for l in range(configs.d_layers)
            ],
            norm_layer=my_Layernorm(configs.d_model),
            projection=nn.Linear(configs.d_model, c_in, bias=True),
        )

        # RevIn
        self.revin = configs.revin
        if self.revin:
            self.revin_layer = RevIN(
                c_in, affine=configs.affine, subtract_last=configs.subtract_last
            )

    def forward(
        self,
        x_enc,
        x_mark_enc,  # past covariates
        x_mark_dec,  # future covariates
        enc_self_mask=None,
        dec_self_mask=None,
        dec_enc_mask=None,
    ):

        # normalise
        if self.revin:
            x_enc = self.revin_layer(x_enc, "norm")

        # decomp init
        mean = torch.mean(x_enc, dim=1).unsqueeze(1).repeat(1, self.pred_len, 1)
        zeros = torch.zeros(
            [x_enc.shape[0], self.pred_len, self.c_in], device=x_enc.device
        )
        seasonal_init, trend_init = self.decomp(x_enc)
        # decoder input
        trend_init = torch.cat([trend_init[:, -self.label_len :, :], mean], dim=1)
        seasonal_init = torch.cat(
            [seasonal_init[:, -self.label_len :, :], zeros], dim=1
        )
        # enc
        enc_out = self.enc_embedding(x_enc, x_mark_enc)
        enc_out, attns = self.encoder(enc_out, attn_mask=enc_self_mask)
        # dec
        dec_out = self.dec_embedding(seasonal_init, x_mark_dec)
        seasonal_part, trend_part = self.decoder(
            dec_out,
            enc_out,
            x_mask=dec_self_mask,
            cross_mask=dec_enc_mask,
            trend=trend_init,
        )
        # final
        dec_out = trend_part + seasonal_part
        dec_out = dec_out[:, -self.pred_len :, :]

        # denorm
        if self.revin:
            dec_out = self.revin_layer(dec_out, "denorm")

        if self.output_attention:
            return dec_out, attns
        else:
            return dec_out  # [B, L, D]

    def train_output(
        self, batch: PaddedData, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        x = batch.data[:, : self.seq_len, :]
        x_pad_mask = batch.pad_mask[:, : self.seq_len]
        y_hat = self(x, x_pad_mask, batch.pad_mask[:, self.seq_len :])
        y = batch.data[:, self.seq_len :, :]

        pad_mask = batch.pad_mask[:, self.seq_len :]
        y_hat = y_hat[pad_mask]
        y = y[pad_mask]

        return y_hat, y

    def generate(
        self, x: torch.Tensor, pad_mask: torch.Tensor, *args, **kwargs
    ) -> Tuple[torch.tensor, torch.tensor]:
        y_hat = self(x, pad_mask, torch.ones(x.shape[0], self.pred_len, self.c_in))
        return y_hat



from abc import ABC, abstractmethod
from typing import Any, Iterable, Tuple

import numpy as np
from pydantic import BaseModel, PositiveInt

import nnts.events

from .utils import Hyperparams


class Evaluator(ABC):

    @abstractmethod
    def evaluate(self) -> None:
        pass


class Forecaster(ABC):
    @abstractmethod
    def forecast(self, data, h: int) -> Any:
        pass


class Trainer(ABC):

    @abstractmethod
    def train(self) -> Evaluator:
        pass

    @abstractmethod
    def create_evaluator(self) -> None:
        pass


class TrainerState(BaseModel):
    epoch: int = 0
    stop: bool = False
    train_loss: float = 0
    valid_loss: float = 0
    best_loss: float = np.inf


class TrainerEvent:
    def __init__(self, state: TrainerState):
        self.state = state


class EpochValidateComplete(TrainerEvent):
    def __init__(self, state: TrainerState):
        super().__init__(state)


class EpochTrainComplete(TrainerEvent):
    def __init__(self, state: TrainerState):
        super().__init__(state)


class EpochBestModel:
    def __init__(self, path: str):
        self.path = path


class EpochTrainer(Trainer):

    def __init__(self, state: TrainerState, params: Hyperparams):
        if state is None:
            state = TrainerState()
        self.state = state
        self.params = params
        self.events = nnts.events.EventManager()

    def before_train(self, train_dl: Iterable) -> None:
        pass

    def train(self, train_dl, valid_dl=None) -> Evaluator:
        self.before_train(train_dl)
        for epoch in range(1, self.params.epochs + 1):
            if self.state.stop:
                break
            self.state.epoch = epoch
            self._train_epoch(train_dl)
            if valid_dl:
                self._validate_epoch(valid_dl)

        return self.create_evaluator()

    def before_train_epoch(self) -> None:
        pass

    def after_train_epoch(self) -> None:
        pass

    def _train_epoch(self, train_dl: Iterable) -> Any:
        if self.params.batches_per_epoch is None:
            self.params.batches_per_epoch = len(train_dl)
        self.before_train_epoch()
        loss = 0
        for i, batch in enumerate(train_dl):
            if i > self.params.batches_per_epoch:
                break
            L = self._train_batch(i, batch)
            loss += L
        loss /= self.params.batches_per_epoch  # len(train_dl
        self.state.train_loss = loss
        self.after_train_epoch()
        self.events.notify(EpochTrainComplete(self.state))
        return loss

    def before_validate_epoch(self) -> None:
        pass

    def after_validate_epoch(self) -> None:
        pass

    def _validate_epoch(self, valid_dl: Iterable) -> Any:
        self.before_validate_epoch()
        loss = 0
        for i, batch in enumerate(valid_dl):
            L = self._validate_batch(i, batch)
            loss += L
        loss /= len(valid_dl)
        self.state.valid_loss = loss
        self.after_validate_epoch()
        self.events.notify(EpochValidateComplete(self.state))
        return loss

    @abstractmethod
    def _train_batch(self, i, batch) -> Any:
        pass

    @abstractmethod
    def _validate_batch(self, i, batch) -> Any:
        pass
